module type arithmetic = sig

  type elt

  (* --- PRINTING ELEMENT --- *)
  val docElt: unit -> elt -> Pretty.doc   (* print element *)
      

  (* --- BOOLEAN OPERATIONS --- *)
  val eq: elt -> elt -> bool              (* equal To *)
  val lt: elt -> elt -> bool              (* Less Than *) 
  val gt: elt -> elt -> bool              (* Greater Than *)
  val leq: elt -> elt -> bool             (* Less Than or Equal To *)
  val geq: elt -> elt -> bool             (* Greater Than or Equal To *)

  (* --- HASHING OPERATION --- *)
  val hashElt: elt -> int                 (* to be used for hashtables *)
  (* should have the property that equal values have the same hash value *)
  (* i.e. forall e1 e2 (if (eq e1 e2) then (hashElt e1 = hashElt e2)) *)
  (* This is not trivial, if for e.g. a floating point implementation
   * considers non-equivalent values to be equal, for e.g. eq (1.000000002, 1.000000003) = true *)

    
  (* --- ELEMENT GENERATION --- *)
  val randElt: unit -> int -> elt (* generates a random element between -int to +int *)
  val zero: elt                   (* The element corresponding to integer 0 *)
  val one: elt                    (* The element corresponding to integer 1 *)
  val int2elt: int -> elt         (* The element corresponding to integer i *)


  (* --- ARITHMETIC OPERATIONS --- *)
  val add: elt -> elt -> elt
  val sub: elt -> elt -> elt
  val mult: elt -> elt -> elt
  val div: elt -> elt -> elt
  val absolute: elt -> elt
  val join: elt -> elt -> elt -> elt

end
