(** This module handles the sel and upd operators *)
module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module IH = Inthash

open Satproc              (* message, etc. *)
open Edag
open Pretty


let arrayid = Util.registerSymbolName "AR"

let debug = ref false 


let log fmt = ignore (E.log "AR: "); E.log fmt
let error fmt = ignore (E.log "AR: "); E.error fmt


(** Remember if we have announced the axiom *)
let haveAnnouncedAxiom = ref false

let announceAxiom () : unit = 
  if not !haveAnnouncedAxiom then begin 
    haveAnnouncedAxiom := true;
    registerUndoRun (fun _ -> haveAnnouncedAxiom := false);
    
    (* SA:3, UA:2, UV:1, UM:0 *)
    let sa = L.Var 3 in 
    let ua = L.Var 2 in 
    let uv = L.Var 1 in 
    let um = L.Var 0 in 
    let upd = L.Fun("upd", [ ua; uv; um ]) in
    let sel = L.Fun("sel", [ sa; upd ]) in
    let sel1 = L.Fun("sel", [ sa; um ]) in 

    let trigger = L.pFun "have" [ sel ]  in 
    let case1 = 
      L.And [ L.eq sa ua; 
              L.eq sel uv ] in 
    let case2 = 
      L.And [ L.ne sa ua; 
              L.eq sel sel1 ]
    in
    let body = 
      (L.All ("sa", 
              L.All ("ua", 
                     L.All ("uv", 
                            L.All ("um", 
                                   L.Pannot(L.ATrigger trigger, 
                                            L.Or [ case1; case2 ])))))) in
    let hyp = Util.registerSymbolName "hyp" in 
    let bodyprf = 
      P.Alli ("sa", 
              P.Alli("ua", 
                     P.Alli("uv", 
                            P.Alli("um", 
                                   P.AnnotI(L.ATrigger trigger,
                                            P.SelUpd (sa, ua, uv, um))))))
    in
    Matching.internForall body bodyprf
  end



let assume (msg: message) : unit = 
  announceAxiom ();
  ()



let init () = 
  haveAnnouncedAxiom := false



(* list of entry points exported as a satproc *)
let entries: Satproc.entries = {
  emptySatproc with 
  init      = init;
  assume    = assume;
  debug     = debug;
  ident     = arrayid;
  descr     = "arrays";
}
