
(* This defines data types for the output of the parser. *)

type axiom = string * string list * Logic.pred list * Logic.pred

type loc = Errormsg.location

(** The input to the prover is presented as a sequence of "inputs".*)
type input =
    (** The next few kinds of inputs can appear only at top level *)
    StartGoal of string * loc           (* The start of a goal. Followed by a 
                                         * number of elements ended by 
                                         * some top-level element *)
  | TopChoose of int * int              (* A choose at the top level *)
  | Axiom  of axiom
  | Import of string
  | Decl   of string

  | Block of (string * string) list * loc
  | UnBlock of loc

  | Eof

  (** The next few kinds of inputs can appear only after a StartGoal **)

  | Prove of Logic.pred * loc (** Prove something. Not a conjunction, 
                                      * nor implication, nor universal 
                                      * quantification. *)
  | ProveInvalid of Logic.pred * loc (** Attempt to prove something that 
                                      * should not be provable, for debugging *)

  | Assume of Logic.pred * loc (** Assume something. Left-hand of an 
                                * implication *)
  | UnAssume of loc     (** retract the last assumption *)

  | Pick of string * loc  (** Pick a variable. Universal quantification *)
  | UnPick  of loc    (** Retract the last variable picked *)

  | Choose of int * int * loc
  | Comment of string * loc
  | EndGoal of loc


val di  : unit -> input -> Pretty.doc

val printLFSyntax : bool ref            (* If true then things are printed in 
                                         * the prefix notation of LF *)


