
(** A satsfiability procedure for some pointer-specific functions *)
module E = Errormsg
module L = Logic
module P = Proof
module U = Util
module IH = Inthash

open Satproc              (* message, etc. *)
open Edag
open Pretty


let ptrid = Util.registerSymbolName "PT"

let debug = ref false 


let log fmt = ignore (E.log "PT: "); E.log fmt
let error fmt = ignore (E.log "PT: "); E.error fmt

let noshareHead = uniqueHead (UNINT (U.registerSymbolName "noshare"))


(* Scan all distinct pairs of fields *)
let rec allFieldPairs (doit: node -> node -> unit) (fl: node list) = 
  match fl with 
    [] -> () (* Done *)
  | f :: fl' -> 
      doit f f;
      List.iter (fun f' -> doit f f') fl';
      allFieldPairs doit fl'

(* Watch for when a noshare node is set to true *)

let null = L.Fun("null", [])

let assume (msg: message) : unit = 
  match msg.body with
    (* noshare *)
  | Intern((L.SetNe|L.SetEq) as bop, n, t) 
    when n.head == noshareHead && t == !trueNode -> 
      if bop = L.SetNe then begin 
        if !debug then 
          ignore (log "proving noshare %a\n" d_message msg);
        (* Make two names *)
        let p1 = L.NVar (newAlphaName "p1") in 
        let p2 = L.NVar (newAlphaName "p2") in
        (* Add a bunch of assumptions about them *)
        let assumvar1 = genNewHypothesisSymbol "PTR" in
        let p = L.And [ L.ne p1 null;
                        L.ne p2 null;
                        L.ne p1 p2; ] 
        in
        newMessageWhy 
          { body = Extern p;
            fromgoal = msg.fromgoal;
            owner = ptrid;
            proof = P.Magic("pointers", p);
          }
          "pointers";
        let assumvar2 = genNewHypothesisSymbol "PTR" in
        let disjuncts: L.pred list ref = ref [] in
        allFieldPairs
          (fun f1 f2 -> 
            let p1f1 = L.Fun("sel", [ p1; f1.exp]) in 
            let p2f2 = L.Fun("sel", [ p2; f2.exp]) in
            disjuncts := 
              L.And [ L.ne p1f1 null; L.eq p1f1 p2f2 ] :: !disjuncts)
          n.args;
        (* Now add the clause *)
        let p = L.Or !disjuncts in 
        Cases.addClauseOfPredicate 
          p
          (P.Magic ("pointers", p))
          ~ingoal:true
        
      end else begin
        if !debug then 
          ignore (log "assuming noshare %a\n" d_message msg);
        
        (* A:1, B:0 *)
        let a = L.Var 1 in 
        let b = L.Var 0 in 
        let getMatchingRule (f1: node) (f2: node) : unit = 
          let sela = L.Fun("sel", [ a; f1.exp ]) in 
          let selb = L.Fun("sel", [ b; f2.exp ]) in 
          let trigger = L.eq sela selb in
          let body = 
            L.All ("a", 
                   L.All ("b", 
                          L.Pannot(L.ATrigger trigger, 
                                   L.Or [ L.eq a null;
                                          L.eq b null;
                                          L.And [ L.ne a null;
                                                  L.ne b null;
                                                  L.eq sela null ];
                                          L.And [ L.ne a null;
                                                  L.ne b null;
                                                  L.ne sela null;
                                                  L.eq a b ] ])))
          in
          let prf = P.Magic ("assume_noshare", body) in
          Matching.internForall body prf
        in
        allFieldPairs getMatchingRule n.args
      end

  | _ -> ()
      



let init () = 
  ()



(* list of entry points exported as a satproc *)
let entries: Satproc.entries = {
  emptySatproc with 
  init      = init;
  assume    = assume;
  debug     = debug;
  ident     = ptrid;
  descr     = "pointers";
}
