/* 
 * Copyright (c) 1999 Cedilla Systems Incorporated.  All Rights Reserved.
 *
 * This software is the confidential and proprietary information of
 * Cedilla Systems Incorporated. ("Confidential Information").  You shall
 * not disclose such Confidential Information and shall use it only in
 * accordance with the terms of the license agreement you entered into
 * with Cedilla Systems.
 * 
 *****************************************************************************/

// #include "main.h"

typedef void* HASH_KEY; /* Notice that the NULL key is reserved to mean an 
                         * empty hash table */

typedef struct HashEntry {
  HASH_KEY key;               /* The key NULL is reserved for empty entries */
  void * data;
} HASH_ENTRY;


typedef int (*key2Hash)(HASH_KEY);
typedef int (*keyCompare)(HASH_KEY, HASH_KEY);


/* Poor's man abstraction of the hash-table internals */
typedef void *PHASH;


/* Create a new hash table give a function to convert a key into a hash 
 * value (an integer) and a comparison function that returns 1 iff the keys 
 * are equal. There can be only one entry with the same key in the hash 
 * table. Note that the NULL key is invalid */
PHASH NewHash(key2Hash k2h, keyCompare kc);

/* Often the hash tables are keyed on strings. */
PHASH NewStringHash(void);
PHASH NewStringPointerHash(void);


void  FreeHash(PHASH);




       /* Looks up the key in the table and returns the corresponding entry. 
        * If the key does not exists, creates a new entry and sets the key. 
        * You must still set the data for this entry. If "found" is not NULL 
        * it sets "*found". */
       /* If you set the key in the resulting entry to NULL, then it is as 
        * if you delete the key from the table */
       /* You can always set the data field of the HASH_ENTRY, but if you 
        * set the key to a non-NULL value it better be equivalent to the 
        * previous one (has the same hash value and compares the same with 
        * other keys). */
HASH_ENTRY *HashLookupOrAdd(PHASH h, HASH_KEY key, int * found);

       /* Returns NULL if the entry is not found. */
HASH_ENTRY *HashLookup(PHASH h, HASH_KEY key);


       /* You can delete an entry by setting its key to NULL. Or you can use 
        * this function if you do not have already looked up the entry. 
        * Returns 1 if the entry actually existed. */
static inline int HashDelete(PHASH h, HASH_KEY key) {
  HASH_ENTRY *e = HashLookup(h, key);
  if(e) {
    e->key = NULL;
    return 1;
  } else
    return 0;
}

        /* Use this function to add an entry, or replace it if already in 
         * the table. Returns 1 if the entry was in the table */
static inline int HashReplace(PHASH h, HASH_KEY key, void *data) {
  int found;
  HASH_ENTRY *e = HashLookupOrAdd(h, key, &found);
  e->data = data;
  return found;
}

                              /* Maps a function to a hash table. The last 
                               * element is a closure. The data is 
                               * overwritten but not placed into another 
                               * bucket ! */
int   MapHash(PHASH, void* (*)(HASH_KEY, void*, void*), void*);

int   IterHash(PHASH, void (*)(HASH_KEY, void*, void*), void*);



                              /* Preallocates some hashes */
int   preallocateHashes(void);
                              /* And release them */
int   releaseHashes(void);
