#include <stdio.h>

/* This is an element of the field on which we are doing the computation. 
 * For now, this should be >= 0 (implicitly since it is unsigned) and < 
 * randPrime */
typedef unsigned long element;

// A value is a vector of elements, one for each run.
typedef element* value;


extern int nrPoints; // How many points we are using 

struct groupNode {
  int g;
  int from;
  struct groupNode* rep;
};

/* The type of a new_value */
typedef struct NewValue {
  int             used; // if this value is used or not

  struct groupNode* group;
  struct NewValue *rep;
  /* If the "rep" pointer is non-null then this value's data is ignored and 
   * is taken from the representative. If "rep" is NULL then this is a 
   * representative. */

  struct NewValue  *func;
  /* If this pointer is non-null then it points to the current value of the 
   * uninterpreted function for this value of the argument */

  element         data[0];
  /* We allocate here the data for the value. We have "nrPoints" elements. */
} NewValue;

/* We have a global array that contains pointers to all the values. */
extern NewValue * *values;
extern int nrUsedValues;   // The index of the first unused value


// Allocates a new value and adds its address to the "values" array
// This function also sets the "used" flag for a value. You just have to
// fill in the data.
NewValue *newValue(void);

// Like malloc, but keeps a count of the space allocated
void* my_malloc(int);

// And release a value
void releaseValue(NewValue *v);


// And release all values (kind of)
static inline void releaseAllValues(void) {
  nrUsedValues = 0;
}


// Get the representative with path compression
static inline NewValue* getValueRep(NewValue *v) {
  NewValue *v1 = v;
  if(! v->rep) return v; // The quick case
  
  while(v1->rep) {
    v1=v1->rep;
  }
  // Go again and reset the representative to v1
  while(v->rep) {
    NewValue *t = v->rep;
    v->rep = v1;
    v = t;
  }
  return v1;
}



typedef enum { VAR_EXP, INTEGER_EXP, SUB_EXP,
	       PLUS_EXP, MINUS_EXP, TIMES_EXP,
	       DIV_EXP, MOD_EXP, FLOOR_EXP,
	       CEIL_EXP, FUN_EXP
} exp_type;

struct explist;

struct exps;

struct exp {
  exp_type tag;
  union {
    char* var;
    int integer;
    struct {
      struct exp *e1;
      struct exp *e2;
    } exps;
    struct {
      char* var;
      struct explist* elist;
    } func;
    struct {
      int i;
      struct exp* e;
    } s;
  } u;
};

struct explist {
  struct exp e;
  struct explist *next;
};


typedef enum { FALSE_ATOM, TRUE_ATOM, EQ_ATOM,
	       NEQ_ATOM, GEQ_ATOM, GT_ATOM,
	       LEQ_ATOM, LT_ATOM, BFUN_ATOM,
	       NOTBFUN_ATOM
} atom_type;

struct atom {
  atom_type tag;
  union {
    struct {
      struct exp e1;
      struct exp e2;
    } s1;
    struct {
      char* str;
      struct explist* elist;
    } s2;
  } u;
};

struct atomlist {
  struct atom a;
  struct atomlist* next;
};

typedef enum { TRUE_PRED, FALSE_PRED, ALL_PRED,
	       EXIST_PRED, AND_PRED, OR_PRED,
	       IMPL_PRED, ATOM_PRED
} pred_type;

struct pred {
  pred_type tag;
  union {
    struct {
      char* var;
      struct pred *p1;
    } s1;
    struct {
      struct pred *p1;
      struct pred *p2;
    } s2;
    struct atom a;
  } u;
};

struct predlist {
  struct pred p;
  struct predlist* next;
};


struct axiom {
  char* str;
  char** vlist;
  struct atomlist* alist;
  struct atom a;
};

typedef enum { GOAL_INPUT, TOPCHOOSE_INPUT, AXIOM_INPUT,
	       IMPORT_INPUT, DECL_INPUT, EOF_INPUT
} input_type;

struct input {
  input_type tag;
  union {
    char* str;
    struct {
      int i1;
      int i2;
    } s1;
    struct axiom a;
  } u;
};

typedef enum { LCONST_LF, LVAR_LF, LFUN_LF,
	       LAPP_LF, LPF_LF
} lf_type;

struct lflist;
struct lfslist;

struct lf {
  lf_type tag;
  union {
    char *str;
    struct {
      struct lfslist* lfsl;
      struct lf* l;
    } s2;
    struct {
      struct lf *l;
      struct lflist *ll;
    } s1;
    struct pred p;
  } u;
};

struct lfslist {
  char* str;
  struct lf l;
  struct lfslist* nect;
};


typedef enum { CONJUNCT_ELEMENT, CHOOSE_ELEMENT, COMMENT_ELEMENT,
	       ASSUME_ELEMENT, RETRACT_ELEMENT, ENDPROOF_ELEMENT
} element_type;

struct element {
  element_type tag;
  union {
    struct atom a;
    struct {
      int i1;
      int i2;
    } s1;
    char *str;
  } u;
};


void print_exp(struct exp* e);      
