
#define bool int

#ifdef DEBUG
// We want to abort because that catches the attention of gdb
#define _ABORT abort()
#define _ASSERT(a) { if(! (a)) { fprintf(stderr, "Assert failed at %s:%d\n", \
                                         __FILE__, __LINE__); _ABORT; }}
#else
#define _ABORT exit(1)
#define _ASSERT(a)
#endif


#define _ERROR0(msg)    { fprintf(stderr, "Error %s:%d: " msg "\n", \
                                  __FILE__, __LINE__); _ABORT; }
#define _ERROR1(msg,a1) { fprintf(stderr, "Error %s:%d: " msg "\n",\
                                  __FILE__, __LINE__, a1); _ABORT; }
#define _ERROR2(msg,a1,a2) { fprintf(stderr, "Error %s:%d: " msg "\n",\
                                  __FILE__, __LINE__, a1, a2); _ABORT; }



extern struct pred *parseGoal(char *filename);


/**** Timing code *****/


/* Run a program and set the "double totalTime" to the number of 
 * milliseconds it took */
#define TIME(totalTime, prog) \
{  struct tms time; \
   clock_t  clk; \
   times(&time); \
   clk=time.tms_utime; \
   { prog; }\
   times(&time); \
   totalTime = (time.tms_utime - clk) / (double)__clock_tick * 1000.0; \
}

/* Run a program for a number of times to total at least the given number of 
 * milliseconds. Set the "int timesRun" variable to how many times it was 
 * run, and the "double totalTime" variable to how many milliseconds it 
 * actually took */
#define TIME_MILLISECONDS(milliseconds, timesRun, totalTime, prog) \
       {timesRun = 1;\
        while(1) {\
           int _i;\
           TIME(totalTime, \
                { for(_i=timesRun;_i>0;_i--){\
                     prog;\
                   }\
                 });\
           /*printf("totalTime=%f ms, timesRun = %d\n", \
                     totalTime, timesRun);*/\
           if(timesRun >= 10000 || totalTime >= milliseconds) {\
              break;\
           }\
           if(totalTime <= 1.0) {\
             timesRun *= 10;\
           } else {\
             int _oldtimes = timesRun;\
             timesRun = (int)(timesRun * (milliseconds / totalTime));\
             if(timesRun == _oldtimes) { timesRun = _oldtimes + 1; }\
           }\
        }\
}


