module E = Errormsg
module L = Logic
module P = Proof
module S = Satproc

module J = Java


(* Conversions from the parameter type t to P.parg and P.proof and back *)
type t = P.proof
let t2p a = a
let p2t a = a

let pa2t (a: P.parg) : P.proof = 
  P.G("just a marker", [ a ])

let t2pa (a: P.proof) : P.parg = 
  match a with 
    P.G("just a marker", [ a' ]) -> a'
  | _ -> P.P a

module A = Prolog

let (&&&) = A.(&&&)


open Pretty

(* New inteface *)
let theProg : A.program = A.emptyProgram false
let prove (prg: A.program) ((a, s) : L.spred) : P.proof option = 
  A.prove prg ~subst:s a


let tcjid = "TCJ"
let debug = ref false



let rec assume acc msg =
  A.debug := if !debug then 1 else 0; J.debug := !debug;
  (* J.debug := !debug; *)
  let goal = msg.S.body in
  (* See if the goal is an atom of interest *)
  let isatom = 
    match goal with
      L.PFun _ | L.Bool _ | L.PNez _ -> true
    | _ -> false 
  in
  let _ = 
    if not isatom then 
      () 
    else if msg.S.goal then begin
      if !debug then
        ignore (E.log "TCJ:***** proving %a@!" L.d_pred goal);
      match prove theProg (goal, L.newSubst 0) with
        None  -> ()
      | Some prf -> begin
          if !debug then 
            ignore (E.log "TCJ: proved goal %a@!" L.d_pred goal);
          raise (S.Proved prf)
      end
    end else begin
      let prf = !S.proveIt msg in
      (* Look for some contradictions *)
      (* checkContra theProg prf a; *)
      (* No contradiction found *)
      A.assume theProg goal prf
        ("ass" ^ msg.S.owner 
         ^ (string_of_int msg.S.id))
    end
  in
  acc
(*
let checkContra prg prf (g: L.pred) = 
  match g with
    L.PFun("candneq", [ L.zero; L.zero ]) ->
      raise (S.Contradiction (P.G("candneqz", [P.P prf])))

  | _ -> ()
*)

let proveIt _ = E.s (E.unimp "TCJ.proveIt")
    

   (* TCJ type constructors *)
let nullary n      = L.Fun(n, [])
let unary n e      = L.Fun(n, [e])
let binary n e1 e2 = L.Fun(n, [e1;e2])
let ternary n e1 e2 e3 = L.Fun(n, [e1;e2;e3])

let nullaryB n      = L.PFun(n, [])
let unaryB n e      = L.PFun(n, [e])
let binaryB n e1 e2 = L.PFun(n, [e1;e2])
let ternaryB n e1 e2 e3 = L.PFun(n, [e1;e2;e3])

let equal e1 e2     = L.eq e1 e2

let oof         = binaryB "of"
let jinstof     = unary "jinstof"
let jelemtype   = unaryB "jelemtype"
let jstatic     = unary "jstatic"
let jarray      = unary "jarray"
let jint        = nullary "jint"
let jlong       = nullary "jlong"
let jbool       = nullary "jbool"
let jchar       = nullary "jchar"
let jshort      = nullary "jshort"
let jbyte       = nullary "jbyte"

let jtable  l    = unary "jtable" l   (* A type of a jump table *)
let jmptable s l = binaryB "jmptable" s l (* An assumption as constructed by 
                                           * VCGen *)
let injmptable e s = binaryB "injmptable" e s

let jinjtab s    =  unary "jinjtab" s  (* A type of a location *)

let jdouble     = nullary "jdouble"
let jfloat      = nullary "jfloat"
let mem         = nullary "mem"
let jclasstype  = nullary "jclasstype"
let jvirtab     = unary "jvirtab"
let jclassdesc  = nullary "jclassdesc"
let implof      = unary "implof"
let jinit       = binary "jinit"
let jimplof     = unary "jimplof"
let initarray   = binaryB "initarray"
let jsubtype    = binaryB "jsubtype"


let ifield c off t  = L.PFun("ifield", [c;off;t])
let vmethod c off s = L.PFun("vmethod", [c;off;s])

let jlObject    = nullary "_4java4lang6Object_vt"
let jiSerializable = nullary "_4java2io12Serializable_vt"
let jlCloneable  = nullary "_4java4lang9Cloneable_vt"
let jlString = nullary "_4java4lang6String_vt"

let ptr a ro t  = L.PFun("ptr", [a;ro;t])
let ptraux base t off ro t' = L.PFun("ptraux", [base;t;off;ro;t'])

let sizeof t sz = L.PFun("jsize", [t; sz])

let ro          = nullary "ro"
let rw          = nullary "rw"

let sel sz = binary ("sel" ^ string_of_int sz)
let upd sz m a e = L.Fun("upd" ^ string_of_int sz, [m;a;e])
let amem i = unary "amem" i
let afpu i = unary "afpu" i

let add    = binary "add"
let band   = binary "band"
let bor    = binary "bor"
let sub    = binary "sub"
let adc    = ternary "adc"
let sbb    = ternary "sbb"
let mul    = binary "mul"
let bsar   = binary "bsar"
let bshr   = binary "bshr"
let bshl   = binary "bshl"

let saferd sz    = unaryB ("saferd" ^ string_of_int sz)
let safewr sz    = binaryB ("safewr" ^ string_of_int sz)
let safebcopy    = ternaryB "safebcopy"
let jsafeintern  = unaryB "jsafeintern"

let ceq   = unaryB "ceq"
let cneq  = unaryB "cneq"
let cb    = unaryB "cb"
let cnb   = unaryB "cnb"
let cbe   = unaryB "cbe"
let cnbe  = unaryB "cnbe"
let clt   = unaryB "clt"
let cnlt  = unaryB "cnlt"
let cle   = unaryB "cle"
let cnle  = unaryB "cnle"


let ult e1 e2     = binaryB "ult" e1 e2
let ge e1 e2      = binaryB "ge" e1 e2
let csubb e1 e2   = binaryB "csubb" e1 e2

let wrd1 e1       = unary "wrd1" e1
let wrd2 e1       = unary "wrd2" e1
let pack2         = binary "pack2"


  (* The base can be a sel4, a variable or a constant *)
  (* The offset can be a numeral optionally + idx * size with size = 4 or 8 *)
let getBaseAddr a base off = L.PFun("getBaseAddr", [a;base;off])
  
   (* Find the type of a base address. Works when base is a variable, a 
    * constant or the result of a memory read. Use splitAddr on an address to 
    * obtain a base type *)
let baseType a t  = binaryB "baseType" a t

let nonnull     = unaryB "nonnull"


let i n         = L.integer n

let rec addVars n acc = 
  if n = 0 then acc
  else addVars (n - 1) (L.UVar (n - 1) :: acc)

let maxUnifId = ref 0
let mkVars n = maxUnifId := n - 1; addVars n []
    

   (* Proof builders *)
let pf: string -> A.program -> L.subst -> t list -> P.proof  = 
  fun name prg s (args: t list) -> P.G(name, List.map t2pa args)

  (* Drop some args *)
let dropArg next = 
  fun name prg s args ->
    match args with 
      [] -> E.s (E.bug "Trying to drop from an empty arg list")
    | _ :: rest -> next name prg s rest

  (* Prepend some args *)
let addArgs (newargs: t list) next = 
  fun name prg s (args: t list) -> 
    let doe e = L.unrollUVarsError s e in
    let rec doParg = function
      | P.E e  -> (P.E (doe e))
      | P.I n  -> P.I n
      | P.P(P.G(n, args)) -> P.P(P.G(n, List.map doParg args))
      | P.F(L.Bool _) as p -> p
      | P.F(L.PNez e) -> P.F(L.PNez (doe e))
      | P.F(L.PFun (n, args)) -> P.F(L.PFun(n, List.map doe args))
      | parg -> E.s (E.bug "addArgs: %a" insert (P.dpa parg))
    in
    let rec doAddArgs = function
        [] -> args
      | parg :: rest -> pa2t (doParg (t2pa parg)) :: doAddArgs rest
    in
    next name prg s (doAddArgs newargs)

let collectPf (acc: t list) (pf: P.proof) = p2t pf :: acc


let simpleAx ~name ~conc ~sub ~mkprf = 
  A.extendProgram theProg
    { A.name = name;
      A.maxUnif = !maxUnifId;
      A.conc = conc;
      A.sub = sub;
      A.init = [];
      A.isassume = false;
      A.mkprf = fun prg s (acc: t list) -> mkprf name prg s (List.rev acc);
    } 

let goal g     = A.sgPROVE g collectPf
let isassume a = A.sgASSUMPTION a collectPf

let sgprint doc = 
  A.sgCHECK (fun _ _ acc -> ignore (fprint !E.logChannel 80 doc); (true, acc))

let unroll e s = L.unrollExp (e, s)
let fstunroll e s = fst (L.unrollExp (e, s))

let isVarOrConst c prg s acc = 
  match fstunroll c s with
    L.Var _ | L.Fun(_, []) -> (true, acc)
  | _ -> (false, acc)


let ptlst () (pal: t list) = 
  docList ~sep:(chr ',' ++ break) P.dpa () 
    (List.map t2pa pal)
let parglst () (pal: P.parg list) = 
  docList ~sep:(chr ',' ++ break) P.dpa () 
    pal

            

(* Add axioms to the program *)

(*** Subclassing ***)
let _ =
  let rec isJavaSubClass t1 t2 prg s acc = 
    if !debug then
      ignore (E.log "TCJ:    ***** isJavaSubClass (@[%a <@? %a@])@!" 
                L.d_sexp (t1, s) L.d_sexp (t2, s));
    match L.unrollUVarsError s t1, L.unrollUVarsError s t2 with
      (L.Fun(n1, []) as cls1), (L.Fun(n2, []) as cls2) -> begin
        if J.checkSubtype n1 n2 then
          (true, p2t (P.G("jsubcls", [ P.E (L.Fun(n1,[]));
                                       P.E (L.Fun(n2,[]))])) :: acc)
        else (false, acc)
      end
(*
    | (L.Fun("jarray", [t1']), L.Fun("jarray", [t2'])) -> begin
        let (res, prf :: _) = isJavaSubtype t1' t2' prg s [] in
        if res then 
          (true, P.P(P.G("jextarr", [ prf ])) :: acc)
        else
          (false, acc)
      end
    | (L.Fun("jarray", [t1']), L.Fun("jarray", [t2'])) -> begin
        let (res, prf :: _) = isJavaSubtype t1' t2' prg s [] in
        if res then 
          (true, P.P(P.G("jextarr", [ prf ])) :: acc)
        else
          (false, acc)
      end
    | (L.Fun("jarray", [t1']), L.Fun("jinstof", [L.Fun(n2, [])])) -> begin
        match n2 with 
          "_4java4lang6Object_C" -> 
            (true, P.P(P.G("jextarrObject", [P.E t1'])) :: acc)
        | "_4java4lang9Cloneable_C" -> 
            (true, P.P(P.G("jextarrCloneable", [P.E t1'])) :: acc)
        | "_4java2io12Serializable_C" -> 
            (true, P.P(P.G("jextarrSerializable", [P.E t1'])) :: acc)
        | _ -> (false, acc)
    end
*)
    | _ -> (false, acc)
  in
  let mkprf (n: string) prg (s: L.subst) = function
      [_; btprf; stypepf] -> pf n prg s [stypepf; btprf]
    | _ -> E.s (E.bug "javaSubtype")
  in
  let [_e; _t; _base; _t'] = mkVars 4 in
  simpleAx "ext"
    (oof _e (jinstof _t))
    (goal (getBaseAddr _e _base (i 0)) &&&
     goal (baseType _base _t') &&&
     goal (jsubtype _t' (jinstof _t)) &&&
     A.sgCUT)
     mkprf;
  simpleAx "extarr"
    (oof _e (jarray _t))
    (goal (getBaseAddr _e _base (i 0)) &&&
     goal (baseType _base (jarray _t')) &&&
     A.sgCOMMIT &&&
     goal (jsubtype _t' _t) &&&
     A.sgCUT)
     mkprf;

  (* Now subtyping *)
  simpleAx "jstcls"
    (jsubtype (jinstof _t) (jinstof _t'))
    (A.sgCUT &&&
     A.sgCHECK (isJavaSubClass _t _t') &&&
     A.sgCUT)
    pf;
  simpleAx "jstarr"
    (jsubtype (jarray _t) (jarray _t'))
    (A.sgCUT &&&
     goal (jsubtype _t _t') &&&
     A.sgCUT)
    pf;
  simpleAx "jstarrObject"
    (jsubtype (jarray _t) (jinstof jlObject))
    (A.sgCUT)
    (addArgs [pa2t (P.E _t)] pf);
  simpleAx "jstarrCloneable"
    (jsubtype (jarray _t) (jinstof jlCloneable))
    (A.sgCUT)
    (addArgs [pa2t (P.E _t)] pf);
  simpleAx "jstarrSerializable"
    (jsubtype (jarray _t) (jinstof jiSerializable))
    (A.sgCUT)
    (addArgs [pa2t (P.E _t)] pf)
  


       

(* Memory safety *)
let _ = 
  (* Automatic conversion from jlong to jint if used in a 4-byte context *)
  let typeConv sz torig t' prg s (acc: t list) = 
    let newt, newacc = 
      match fstunroll torig s, acc with
        L.Fun("jlong", []), ptrpf :: rest when sz = 4 -> 
          jint, p2t (P.G("longintl", [t2pa ptrpf])) :: rest
      | _ -> torig, acc
    in
    (A.eunify prg (t',s) (newt, s), newacc)
  in
  let [_a;_t;_e;_m;_t'] = mkVars 5 in
  let memAxioms sz =
    simpleAx ("rd" ^ (string_of_int sz))
      (saferd sz _a)
      (A.sgIsGround _a &&&
       A.sgCUT &&&
       goal (ptr _a _e _t) &&&
       A.sgCUT &&&
       A.sgCHECK (typeConv sz _t _t') &&&
       goal (sizeof _t' (i sz)) &&&
       A.sgCUT)
      pf;
    simpleAx ("wr" ^ (string_of_int sz))
      (safewr sz _a _e)
      (A.sgIsGround _a &&&
       A.sgIsGround _e &&&
       goal (ptr _a rw _t) &&&
       A.sgCUT &&&
       A.sgCHECK (typeConv sz _t _t') &&&
       goal (sizeof _t' (i sz)) &&&
       A.sgCUT &&&
       goal (oof _e _t') &&&
       A.sgCUT)
      pf;
    simpleAx ("basety" ^ (string_of_int sz))
      (baseType (sel sz _m _a) _t)
      (A.sgIsGround _m &&&
       A.sgIsGround _a &&&
       goal (ptr _a _e _t) &&&
       A.sgCUT &&&
       A.sgCHECK (typeConv sz _t _t') &&&
       goal (sizeof _t' (i sz)) &&&
       goal (oof _m mem))
      (fun _ -> pf ("ty" ^ string_of_int sz));
    simpleAx ("ty" ^ string_of_int sz)
      (oof (sel sz _m _a) _t)
      (A.sgIsGround _m &&&
       A.sgIsGround _a &&&
       goal (ptr _a _e _t) &&&
       A.sgCHECK (typeConv sz _t _t') &&&
       A.sgCUT &&&
       goal (sizeof _t' (i sz)) &&&
       goal (oof _m mem))
      pf;
    simpleAx ("ofmem" ^ (string_of_int sz))
      (oof (upd sz _m _a _e) mem)
      (A.sgIsGround _a &&&
       A.sgIsGround _e &&&
       A.sgIsGround _m &&&
       A.sgCUT &&&
       goal (ptr _a rw _t) &&&
       A.sgCUT &&&
       A.sgCHECK (typeConv sz _t _t') &&&
       goal (sizeof _t' (i sz)) &&&
       goal (oof _e _t') &&&
       goal (oof _m mem))
      pf
  in
  memAxioms 4;
  memAxioms 8;
  simpleAx "ofamem"
    (oof (amem _m) mem)
    (A.sgCUT &&&
     A.sgCHECK (fun prg s acc -> (true, [ pa2t (P.E 
                                                (L.unrollUVarsError s _m))])))
    pf


(* Conversions from long to int. Put first so that it is tried last *)
let _ =
  let [_e; _ro] = mkVars 2 in
  simpleAx "longinth"
    (ptr _e _ro jint)
    (goal (ptr (add _e (i (-4))) _ro jlong))
    pf


(* Find out the pointer types *)
let _ = 
  let [_a; _t; _ro; _base; _idx; _size; _off; _t'] = mkVars 8 in
  let mkptrpf (eqpf: P.parg) name (args: P.parg list) : P.proof = 
    let p = P.G(name, args) in
    match eqpf with
      P.P(P.G("eqid", _)) -> p
    | _ -> P.G("ptr~", [eqpf; P.P p])
  in
  let combineEqProofs (p1: P.parg) (p2: P.parg) : P.parg = 
    match p1, p2 with
      P.P(P.G("eqid", _)), _ -> p2
    | _, P.P(P.G("eqId", _)) -> p1
    | P.P(P.G("eqCheck",[P.E e2; P.E e1])), 
        P.P(P.G("eqCheck",[P.E e3; P.E e2'])) when e2 = e2' ->
          if e1 = e3 then
            P.P(P.G("eqid", [P.E e1]))
          else
            P.P(P.G("eqCheck", [P.E e3; P.E e1]))
    | _ -> E.s (E.bug "combineEqProofs @[%a@]@!" ptlst [pa2t p1; pa2t p2])
  in
  let ptrpf n _ _ (acc: t list) : P.proof = 
    let acc = List.map t2pa acc in
    match acc with 
    | [eqpf; P.P(P.G("jconst", [e;P.E(L.Fun("jstatic", [t]))])); 
       P.P(P.G("rdstatic", []))] 
      -> mkptrpf eqpf "staticptr" [e; P.E t]
          
    | [eqpf; P.P(P.G("jconst", [e; P.E(L.Fun("jinit", [t; len]))])); 
       P.P(P.G("rdinit", []))] 
      -> mkptrpf eqpf "initElem" [P.P(P.G("jinit", [e; P.E t; P.E len]))]
          
    | [eqpf; iprf; P.P(P.G("ptrField", [fldpf; nonnullpf]))]
      -> mkptrpf eqpf "instFld" [fldpf; iprf; nonnullpf]
          
    | [eqpf; aprf; P.P(P.G("arrLen", [nonnullpf]))]
      -> mkptrpf eqpf "arrLen" [aprf; nonnullpf]

    | [eqpf; aprf; P.P(P.G("arrElem", [ P.I _; szpf; 
                                        nonnullpf; ultpf; mempf]))] 
      -> mkptrpf eqpf "arrElem" [aprf; mempf; nonnullpf; szpf; ultpf]

    | [eqpf; aprf; P.P(P.G("arrInt", [ szpf; eqintprf; 
                                       nonnullpf; ultpf; mempf]))] 
      -> mkptrpf (combineEqProofs eqpf eqintprf) 
          "arrElem" [ aprf; mempf;  
                      nonnullpf; szpf; ultpf]

    | [eqpf; iprf; P.P(P.G("ptrauxvirtab", [nonnullpf]))] 
      -> mkptrpf eqpf "ptrvirtab" [iprf; nonnullpf]

    | [eqpf; aprf; P.P(P.G("ptrauxvirtabarr", [nonnullpf]))] 
      -> mkptrpf eqpf "ptrvirtabarr" [aprf; nonnullpf]

    | [eqpf; vtabprf; P.P(P.G("implvmethod", [vmprf]))]
      -> mkptrpf eqpf "ptrvmethod" [vmprf; vtabprf] 

    | [eqpf; jmptablpf; P.P(P.G("fetchJmpTable", [ultpf]))] 
        -> mkptrpf eqpf "jmpTabElem" [jmptablpf; ultpf]
 
    | [eqpf; jmptablpf; P.P(P.G("fetchJmpTableImm", [offeqprf; ultpf]))] 
        -> mkptrpf (combineEqProofs eqpf offeqprf)
            "jmpTabElem" [jmptablpf; ultpf]
 
    | acc -> ignore (E.log "ptrpf: @[%a@]@!" parglst acc); P.G(n, acc)
  in
  simpleAx "doptr"
    (ptr _a _ro _t)
    (A.sgIsGround _a &&&
     goal (getBaseAddr _a _base _off) &&&
     goal (baseType _base _t') &&&
     A.sgCOMMIT &&&
     goal (ptraux _base _t' _off _ro _t) &&&
     A.sgCOMMIT)
    ptrpf


(**** Get the base address ***)

(* The base if a var, a constant or a sel4. The offset is an integer added 
 * (optionally) to an index * [4|8] *)

(* the proof of a subgoal (getBaseAddr _addr _base _off) is a proof of 
 * (= (add _base _off) _addr), where the (add .. ..) is canonized to 
 * eliminate addition by zero *)

   (* Construct the proof of a split *)
let splitProof e s (b, bs) (off, offs) = 
  let esplit = (* Construct the canonical representation of an expression *)
    let base = L.unrollUVarsError bs b in
    match off with
    | z when z = L.zero -> base
    | L.Fun("add", [idx; z]) when z = L.zero -> 
        add base (L.unrollUVarsError offs idx)
    | _ -> add base (L.unrollUVarsError offs off)
    in
  let e' = L.unrollUVarsError s e in
  if e' = esplit then
    P.G("eqid", [P.E e'])
  else
    P.G("eqCheck", [P.E esplit; P.E e'])
      
let _ = 
  let rec extractBase a base off = 
    fun prg s acc ->
      try
        match doExtract (a,s) with
        | Some (Some bs, os) -> 
            let res = 
              A.eunify prg (base, s) bs &&
              A.eunify prg (off, s) os in
            (res, 
             if res then p2t (splitProof a s bs os) :: acc else acc)
        | _ -> (false, acc)
      with Not_found -> (false, acc)

  (* The base, if present, is always a var, const or a sel4. The option is 
   * always (i off) or (add (mul idx (i f)) (i off)) with f = 4 or 8 *)
  and doExtract (ads : L.sexp) : ((L.sexp option * L.sexp) option)= 
    let (a, s) = L.unrollExp ads in
    match a with 
      (L.Fun(_, []) | L.Var _ | L.Fun("sel4",_)) -> 
        Some (Some (a,s), (i 0, L.emptySubst))
    | L.Int i -> Some (None, (L.Int i, L.emptySubst))
    | L.Fun("add", [e1; e2]) -> begin
        let e1bo = doExtract (e1, s) in
        let e2bo  = doExtract (e2, s) in
        match e1bo, e2bo with
        | Some (Some bs, off1), Some (None, off2) -> 
            Some (Some bs, addOffset off1 off2)
        | Some (None, off1), Some (Some bs, off2) -> 
            Some (Some bs, addOffset off1 off2) 
        | Some (None, off1), Some (None, off2) -> 
            Some (None, addOffset off1 off2)
        | Some (Some _, _), Some (Some _, _) -> None
        | None, _ -> None
        | _, None -> None
    end
    | L.Fun("mul", [e1; e2]) -> begin
        let (e1',s1) = L.unrollExp (e1, s) in
        let (e2',s2) = L.unrollExp (e2, s) in
        match e1', e2' with
        | L.Int f1, L.Int f2 -> 
            Some (None, (L.integer (L.to_int f1 * L.to_int f2), L.emptySubst))
        | L.Int f, _ when L.to_int f = 4 || L.to_int f = 8 -> 
            Some (None, (add (mul e2' (i (L.to_int f))) (i 0), s2))
        | _, L.Int f when L.to_int f = 4 || L.to_int f = 8 -> 
            Some (None, (add (mul e1' (i (L.to_int f))) (i 0), s1))
        | _, _ -> None
    end
    | _ -> None
          
  and addOffset off1s off2s = 
    let (off1, s1) = L.unrollExp off1s in
    let (off2, s2) = L.unrollExp off2s in
    match off1, off2 with
      L.Int o1, L.Int o2 -> 
        (L.integer (L.to_int o1 + L.to_int o2), L.emptySubst)
    | L.Int o1, L.Fun("add", [ L.Fun("mul", [idx2; L.Int f2]); L.Int o2]) -> 
        add (mul idx2 (i (L.to_int f2))) 
          (i (L.to_int o1 + L.to_int o2)), s2
    | L.Fun("add", [ L.Fun("mul", [idx1; L.Int f1]); L.Int o1]), L.Int o2  -> 
        add (mul idx1 (i (L.to_int f1))) 
          (i (L.to_int o1 + L.to_int o2)), s1
    | L.Fun("add", [ L.Fun("mul", [idx1; L.Int f1]); L.Int o1]), 
      L.Fun("add", [ L.Fun("mul", [idx2; L.Int f2]); L.Int o2]) -> 
        let f1 = L.to_int f1 in
        let f2 = L.to_int f2 in
        let o1 = L.to_int o1 in
        let o2 = L.to_int o2 in
        let newf = if f1 = f2 then f1 else raise Not_found in
        let news = 
          if s1 == s2 then s1 else
          if s1 == L.emptySubst then s2 else
          if s2 == L.emptySubst then s1 else
          raise Not_found in
        add (mul (add idx1 idx2) (i newf)) (i (o1 + o2)), news
    | _ -> raise Not_found

  in
  let [_a; _base; _off] = mkVars 3 in
  simpleAx "getBase"
    (getBaseAddr _a _base _off)
    (A.sgCUT &&&
     A.sgCHECK (extractBase _a _base _off))
    (fun _ _ _ [ p ] -> t2p p)

(*** Finding the type of a base address ***)


let _ =                              (* Maybe it is a Java constant *)
  let isJavaConstant a t prg s (acc: t list) = 
    let e', s' = L.unrollExp (a, s) in
    match e' with
      L.Fun(n, []) -> begin
        match J.findTypeConstant n with
          Some (jt, _) -> begin
            let res = A.eunify prg (t,s) (jt, L.emptySubst) in
            (res, p2t (P.G("jconst",[P.E e'; P.E jt])) :: acc)
          end
        | _ -> (false, acc)
      end
    | _ -> (false, acc)
  in
  let mkprf _ _ _ = function
      [ prf ] -> t2p prf
    | acc -> E.s (E.bug "mkprf: isJavaConstant %a@!" ptlst acc)
  in
  let [_a; _t; _len] = mkVars 3 in
  simpleAx "baseTypeAssJava"
    (baseType _a _t)
    (A.sgCHECK (isVarOrConst _a)  &&&
     A.sgCUT &&&
     A.sgOR [ isassume (oof _a _t);       (* Do not cut after the assumption 
                                         * since there might be more that 
                                         * match  *)
            A.sgIsConst _a &&& A.sgCUT &&& A.sgCHECK (isJavaConstant _a _t)])
    mkprf;
  simpleAx "ofclasstype"
    (oof _a jclasstype)
    (A.sgIsConst _a &&&
     A.sgCUT &&&
     A.sgCHECK (isJavaConstant _a jclasstype) &&&
     A.sgCHECK (fun prg s acc -> 
       (true, pa2t (P.E (L.unrollUVarsError s _a)) :: acc)))
    (fun _ _ _ [_; e] -> P.G("jisclasstype", [t2pa e]));
  simpleAx "startInit"
    (initarray _a _t)
    (A.sgIsConst _a &&&
     A.sgCUT &&&
     A.sgCHECK (isJavaConstant _a (jinit _t _len)))
    (dropArg 
       (addArgs [p2t (P.G("jinit", [ P.E _a; P.E _t; P.E _len]))] 
          pf)) ;
  let [_a; _start; _bytes; _t; _len; _m; _sz] = mkVars 7 in
  simpleAx "bcopy"
    (safebcopy (add _a (i J.arrayDataOffset)) _start _bytes)
    (A.sgCHECK (isVarOrConst _a) &&&
     A.sgCHECK (isVarOrConst _start) &&&
     A.sgCUT &&&
     A.sgCHECK (isJavaConstant _start (jinit _t _len)) &&&
     A.sgIsGround _t &&&
     goal (sizeof _t _sz) &&&
     A.sgCHECK (fun prg sub acc -> 
       match fstunroll _bytes  sub, 
             fstunroll _len sub, fstunroll _sz sub with
         L.Int bytes, L.Int len, L.Int sz -> 
           let bytes = L.to_int bytes in
           let len = L.to_int len in
           let sz = L.to_int sz in
           if bytes = len * sz then
             (true, p2t (P.G("eqCheck", [ P.E (L.integer bytes); 
                                         P.E (L.times
                                                (L.integer len)
                                                (L.integer sz))]))
                    :: acc)
           else
             (false, acc)
       | _ -> (false, acc)) &&&
     A.sgCUT &&&
     goal (oof _a (jarray _t)) &&&
     A.sgCUT &&&
     isassume (equal (sel 4 _m (add _a (i J.arrayLenOffset))) _len) &&&
     goal (oof _m mem))
    (dropArg 
       (addArgs [p2t (P.G("jinit", [ P.E _start; P.E _t; P.E _len]))] 
          pf)) ;
  let [_s] = mkVars 1 in
  simpleAx "jinternstring"
    (jsafeintern _s)
    (goal (oof _s (jinstof jlString)) &&&
     A.sgCUT)
    pf ;
  let [_s; _t; _len] = mkVars 3 in
  simpleAx "jinterninit"
    (jsafeintern _s)
    (A.sgCHECK (isJavaConstant _s (jinit _t _len)) &&&
     A.sgCUT)
    (dropArg
       (addArgs [p2t (P.G("jinitS", [ P.E _s ]))]
	  pf))

    

(***** JELEMTYPE ****)
let _ = 
  let [_t] = mkVars 1 in
  simpleAx "jetypeinst"
    (jelemtype _t)
    (goal (oof _t jclasstype))
    pf;
  List.iter 
    (fun chr -> 
      simpleAx ("jetype" ^ String.make 1 chr)
        (jelemtype (i (int_of_char chr)))
        A.sgCUT
        pf)
    ['B'; 'C'; 'D'; 'F'; 'I'; 'J'; 'S'; 'Z'] 
    




(***** PTR AUX ********)
    
let _ =
  let isJavaField c off tfield prg s (acc: t list) = 
    match fstunroll c s, fstunroll off s with
      (L.Fun(cn, []) as c'), L.Int off when 
               L.to_int off >= J.objectOverhead -> begin
        let off = L.to_int off in
        match J.getFieldType cn off with
          None -> (false, acc)
        | Some t -> 
            (A.eunify prg (tfield, s) (t, L.emptySubst), 
             p2t (P.G("jifieldpf",[P.E c'; P.I off; P.E t])) :: acc)
      end
    | _ -> (false, acc)
  in
  let [_base;_c;_off;_tfield] = mkVars 4 in
  simpleAx "ptrField"
    (ptraux _base (jinstof _c) _off rw _tfield)
    (A.sgCHECK (isJavaField _c _off _tfield) &&&
     A.sgCUT &&&
     goal (nonnull _base))
    pf
       

let _ = 
  let computeIndex base off sz idx prg s (acc: t list) = 
    match fstunroll off s, fstunroll sz s with
      L.Int off, L.Int sz when L.to_int sz >= 4 && 
                               L.to_int off >= J.arrayDataOffset
	&& (L.to_int off - J.arrayDataOffset) mod (L.to_int sz) = 0 ->
        let off = L.to_int off in
        let sz = L.to_int sz in
        let iidx = (off - J.arrayDataOffset) / sz in
        let res = A.eunify prg (idx, s) (i iidx, L.emptySubst) in
        let base' = L.unrollUVarsError s base in
        let eqprf = 
          p2t (splitProof (add base' (i off)) L.emptySubst
                          (base', L.emptySubst)
                          (add (mul (i iidx) (i sz)) (i J.arrayDataOffset),
			   L.emptySubst)) in
        (res, if res then eqprf :: acc else acc) 
    | _ -> (false, acc)
  in
  let [_base;_t;_idx;_m;_sz;_off] = mkVars 6 in
  let mkArray sz = 
    simpleAx "arrElem"
      (ptraux _base (jarray _t)
	 (add (mul _idx (i sz)) (i J.arrayDataOffset)) rw _t)
      (A.sgIsGround _base &&&
       A.sgIsGround _t &&&
       goal (sizeof _t (i sz)) &&&
       A.sgCUT &&&
       goal (nonnull _base) &&&
       A.sgCUT &&&   (* !!!! m is not ground here *)
       goal (ult _idx (sel 4 _m (add _base (i J.arrayLenOffset)))) &&&
       goal (oof _m mem))
      (fun n _ _ acc -> P.G(n, P.I sz :: List.map t2pa acc))
  in
  let mkArrayLen sz = 
    simpleAx "arrElemLen"
      (ptraux _base (jarray _t)
	 (add (mul _idx (i sz)) (i J.arrayDataOffset)) rw _t)
      (A.sgIsGround _base &&&
       A.sgIsGround _t &&&
       goal (sizeof _t (i sz)) &&&
       A.sgCUT &&&
       goal (nonnull _base) &&&
       A.sgCUT &&&   (* !!!! m is not ground here *)
       goal (ult _idx (sel 4 _m (add _base (i J.arrayLenOffset)))) &&&
       goal (oof _m mem))
      (fun n _ _ acc -> P.G(n, P.I sz :: List.map t2pa acc))
  in
  mkArray 4;
  mkArray 8;
  simpleAx "arrInt"
    (ptraux _base (jarray _t) _off rw _t)
    (A.sgIsGround _base &&&
     A.sgIsGround _t &&&
     A.sgIsInt _off (fun off -> off >= J.arrayDataOffset && off mod 4 = 0) &&&
     goal (sizeof _t _sz) &&&   (* _sz gets instantiated here *)
     A.sgCUT &&&
     A.sgCHECK (computeIndex _base _off _sz _idx) &&&
     goal (nonnull _base) &&&
     A.sgCUT &&&   (* !!!! m is not ground here *)
     goal (ult _idx (sel 4 _m (add _base (i J.arrayLenOffset)))) &&&
     goal (oof _m mem))
    pf

          
let _ =
  let [_base; _t] = mkVars 2 in
  simpleAx "arrLen"
    (ptraux _base (jarray _t) (i J.arrayLenOffset) ro jint)
    (A.sgIsGround _base &&&
     A.sgCUT &&&
     goal (nonnull _base))
    pf

let _ = 
  let [_base; _t] = mkVars 2 in
  simpleAx "rdstatic"
    (ptraux _base (jstatic _t) (i 0) rw _t)
    (A.sgIsGround _base)
    pf
       

let _ = 
  let [_base; _c] = mkVars 2 in
  simpleAx "ptrauxvirtab"
    (ptraux _base (jinstof _c) (i 0) ro (jvirtab _c))
    (A.sgIsGround _c   &&&
     A.sgCUT &&&
     goal (nonnull _base))
    pf

let _ = 
  let [_base; _t] = mkVars 2 in
  simpleAx "ptrauxvirtabarr"
    (ptraux _base (jarray _t) (i 0) ro (jvirtab jlObject))
    (A.sgIsGround _t &&&
     A.sgCUT &&&
     goal (nonnull _base))
    pf
       
let _ = 
  let isJavaVirtualMethod c off sg prg s (acc: t list) = 
    match L.unrollUVarsError s c, L.unrollUVarsError s off with
      (L.Fun(cn, []) as cls), L.Int off
	when L.to_int off >= J.vtOverhead && (L.to_int off) land 3 = 0 
      -> begin
        let off = L.to_int off in
        let result, sgconst = 
          match L.unrollExp (sg, s) with
            (L.UVar i, s') when s'.(i) = L.Unif -> 
            (* We are actually inquiring about the signature *)
              begin 
                match J.getSignatureVirtual cn off with
                  Some sgname -> begin
                    let sgconst = L.Fun(sgname, []) in
                    let res = A.eunify prg (sgconst, L.emptySubst) (sg, s) in
                    assert(res);
                    res, sgconst
                  end
                | None -> false, L.zero
              end
          | (L.Fun(sgname, []) as sgconst, _) -> (* We are just checking *)
              J.checkSignatureVirtual cn off sgname, sgconst
          | _ -> E.s (E.bug "Unexpected signature %a" L.d_sexp (sg, s))
        in
        if result then
          (true, p2t (P.G("jvirmethod", 
                         [P.E cls; P.I off; P.E sgconst])) :: acc)
        else
          (false, acc)
      end
    | _ -> (false, acc)
  in
  let [_base;_c;_off;_sg] = mkVars 4 in
  simpleAx "implvmethod"
    (ptraux _base (jvirtab _c) _off ro (jimplof _sg))
    (A.sgIsGround _c  &&&
     A.sgCUT &&&
     A.sgCHECK (isJavaVirtualMethod _c _off _sg))
    pf


let _ = 
  let [_base; _t] = mkVars 2 in
  simpleAx "rdinit"
    (ptraux _base (jinit _t (i 1)) (i 0) ro _t)
    (A.sgIsGround _t &&&
     A.sgCUT)
    pf
        
       
(**** Size of ****)
let _ = 
  let computeSize t sz prg s (acc: t list) = 
    let setSize n pname pargs = 
      if A.eunify prg (sz, s) (i n, L.emptySubst) then
        (true, p2t (P.G(pname, List.map (fun e -> P.E e) pargs)) :: acc)
      else
        (false, acc)
    in
    match L.unrollUVarsError s t with
      L.Fun(n1, args) -> begin
        match n1 with
          "jint" -> setSize 4 "szint" []
        | "jdouble" -> setSize 8 "szdouble" []
        | "jfloat" -> setSize 4 "szfloat" []
        | "jinstof" -> setSize 4 "szjinst" args
        | "jlong" -> setSize 8 "szlong" []
        | "jarray" -> setSize 4 "szjarr" args
        | "jbool" -> setSize 4 "szbool" []
        | "jshort" -> setSize 4 "szshort" []
        | "jbyte" -> setSize 4 "szbyte" []
        | "jchar" -> setSize 4 "szchar" []
        | "jvirtab" -> setSize 4 "szvirtab" args
        | "jimplof" -> setSize 4 "szjimplof" args
        | "jexchandler" -> setSize 4 "szjexchandler" []
        | "jinjtab" -> setSize 4 "szjinjtab" args
        | _ -> (false, acc)
      end
    | _ -> (false, acc)
  in
  let sizepf _ _ _ (acc: t list) = 
    let acc = List.map t2pa acc in
    match acc with 
      P.P (szpf) :: rest -> szpf
    | acc -> E.s (E.bug "Cannot create proof of sizeof: %a" parglst acc)
  in
  let [_t; _sz] = mkVars 2 in
  simpleAx "sizeof"
    (sizeof _t _sz)
    (A.sgCUT &&&
     A.sgCHECK (computeSize _t _sz))
    sizepf



(*** NULL CHECK ****)
let _ =
  let [_e] = mkVars 1 in
  simpleAx "nullcandne"
    (nonnull _e)
    (isassume (cneq (band _e _e)) &&&
     A.sgCUT)
    pf;
  simpleAx "nullcsubne"
    (nonnull _e)
    (isassume (cneq (sub _e (i 0))) &&&
     A.sgCUT)
    pf
    


(*** BOUNDS CHECKING ***)
let _ =
  let intUlt i' len prg s (acc: t list) = 
    match fstunroll i' s, fstunroll len s with
      L.Int i', L.Int len when L.to_int i' >= 0 && 
                               L.to_int i' < L.to_int len -> 
        let len = L.to_int len in
        let i' = L.to_int i' in
        (true, 
         p2t (P.G("geqCheck", [P.E (i len); P.E (add (i i') (i 1))])) :: 
         p2t (P.G("geqCheck", [P.E (i i'); P.E (i 0)])) :: acc)
    | _ -> begin 
        (false, acc)
    end
  in
  let [_i; _m; _a; _len] = mkVars 4 in
  simpleAx "ultinit"
    (ult _i (sel 4 _m (add _a (i 4))))
    (isassume (equal (sel 4 _m (add _a (i 4))) _len) &&&
     A.sgCUT &&&
     A.sgCHECK (intUlt _i _len))
    pf;
  let [_e1; _e2] = mkVars 2 in
  simpleAx "ultcsubb"
    (ult _e1 _e2)
    (isassume (cb (sub _e1 _e2)) &&& 
     A.sgCUT)
    pf;
  simpleAx "ultcsubnbe"
    (ult _e1 _e2)
    (isassume (cnbe (sub _e2 _e1)) &&&
     A.sgCUT)
    pf;
  simpleAx "gecandnlt"
    (ge _e1 (i 0))
    (isassume (cnlt (band _e1 _e1)))
    pf;
  simpleAx "gecsubnlt"
    (ge _e1 (i 0))
    (isassume (cnlt (sub _e1 (i 0))))
    pf;
  simpleAx "ultzero"
    (ult (i 0) _i)
    (isassume (cnbe (band _i _i)) &&&
     A.sgCUT)
    pf;

  List.iter 
    (fun (n, op) -> 
      simpleAx n
	(ge (op _e1 _e2) (i 0))
	(isassume (cnlt (op _e1 _e2)) &&&
	 A.sgCUT)
	pf)
    [("genltadd", add); ("genltsub", sub);
      ("genltbsar", bsar); ("genltbshr", bshr); ("genltbshl", bshl); 
      ("genltband", band); ("genltbor", bor)]

(*  simpleAx "genltadd"
    (ge (add _e1 _e2) (i 0))
    (isassume (cnlt (add _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltsub"
    (ge (sub _e1 _e2) (i 0))
    (isassume (cnlt (sub _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltbsar"
    (ge (bsar _e1 _e2) (i 0))
    (isassume (cnlt (bsar _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltbshr"
    (ge (bshr _e1 _e2) (i 0))
    (isassume (cnlt (bshr _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltbshl"
    (ge (bshl _e1 _e2) (i 0))
    (isassume (cnlt (bshl _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltband"
    (ge (band _e1 _e2) (i 0))
    (isassume (cnlt (band _e1 _e2)) &&&
     A.sgCUT)
    pf;
  simpleAx "genltbor"
    (ge (bor _e1 _e2) (i 0))
    (isassume (cnlt (bor _e1 _e2)) &&&
     A.sgCUT)
    pf
*)  
    


(***** Virtual methods ****)
let implSpec e sg = L.PFun("implSpec", [e;sg]) 
   
let _ =
  let [_e; _sg] = mkVars 2 in
  simpleAx "implspec"
    (implSpec _e _sg)
    (A.sgIsGround _e &&&
     A.sgIsGround _sg &&&
     goal (baseType _e (jimplof _sg)))
    pf

        
(****** Floating point ****)
let tdouble = nullary "tdouble"
let _ = 
  let [_e1; _e2] = mkVars 2 in
  let addFloatBinary bin = 
    simpleAx (bin ^ "d")
      (oof (L.Fun(bin, [_e1; _e2])) tdouble)
      (A.sgCUT &&&
       goal (oof _e1 tdouble) &&&
       A.sgCUT &&&
       goal (oof _e2 tdouble) &&&
       A.sgCUT)
      pf
  in
  let bins = ["fadd"; "fsub"; "fdiv"; "fmul"] in
  List.iter addFloatBinary bins;
  let addFloatUnary un = 
    simpleAx (un ^ "d") 
      (oof (L.Fun(un, [_e1])) tdouble)
      (A.sgCUT &&&
       goal (oof _e1 tdouble) &&&
       A.sgCUT)
      pf
  in
  let uns = ["fchs";"fcos";"fsin"] in
  List.iter addFloatUnary uns;
  let addConvert (fname, fty) (tname, tty) = 
    match tty with
      None -> ()
    | Some tty -> 
        let subgoal = match fty with
          None -> A.sgCUT
        | Some fty -> A.sgCUT &&& goal (oof _e1 fty)
        in
        simpleAx (fname ^ "2" ^ tname)
          (oof (L.Fun(fname ^ "_" ^ tname, [_e1])) tty)
          subgoal
          pf
  in
  let sources = [ ("i32", Some jint); 
                  ("f32", Some jfloat); ("f64", Some jdouble)] in
  List.iter (fun s -> addConvert s ("fp", Some tdouble)) sources;
  List.iter (fun t -> addConvert ("fp", Some tdouble) t) sources;
  simpleAx "packdouble"
    (oof (pack2 (wrd1 _e1) (wrd2 _e1)) jdouble)
    (A.sgCUT &&&
     goal (oof _e1 jdouble))
    pf

let _ = 
  let [_i] = mkVars 1 in
  simpleAx "ofafpu"
    (oof (afpu _i) tdouble)
    A.sgCUT
    (addArgs [pa2t (P.E _i)] pf)
  


(********* Booleans ******)
(*
let _ = 
  let [] = mkVars 0 in
  let addBool n = 
    simpleAx ("ofbool" ^ string_of_int n)
      (oof (i n) jbool)
      A.sgCUT
      pf
  in
  addBool 0; addBool 1
*)      
(**** Anything is a jint, jshort, jbyte, jchar, jbool !!!  ****)
let _ = 
  let [_e] = mkVars 1 in
  let addOne (n, t) = 
    simpleAx (n ^ "any")
      (oof _e t)
      (A.sgIsGround _e &&&
       A.sgCUT)
      (addArgs [pa2t (P.E _e)] pf)
  in
  List.iter addOne [ ("jbool", jbool); 
                     ("jint", jint); ("jbyte", jbyte); ("jchar", jchar);
                     ("jshort", jshort)]


(**** NULL ****)
let _ = 
  let [_e; _c] = mkVars 2 in
  simpleAx "of0ieq"
    (oof _e (jinstof _c))
    (A.sgIsGround _e &&&
     A.sgIsGround _c &&&
     goal (equal _e (i 0)))
    (addArgs [pa2t (P.E _c)] pf);
  simpleAx "of0aeq"
    (oof _e (jarray _c))
    (A.sgIsGround _e &&&
     A.sgIsGround _c &&&
     goal (equal _e (i 0)))
    (addArgs [pa2t (P.E _c)] pf)

let _ = 
  let [_e] = mkVars 1 in
  let types = [ ("f", jfloat, fun _ -> []); (* ("d", jdouble); *)
                ("i", jinstof _e, fun s -> 
                  [pa2t (P.E (L.unrollUVarsError s _e))]); 
                ("a", jarray _e, fun s -> 
                  [pa2t (P.E (L.unrollUVarsError s _e))])] in
  List.iter (fun (n, t, mkarg) -> 
    simpleAx ("of0" ^ n)
      (oof (i 0) t)
      (A.sgCUT &&&
       A.sgCHECK (fun prg s acc -> (true, mkarg s)))
      pf)
    types

     
(*
let _ = 
  let [_ro; _t] = mkVars 2 in
  simpleAx "ptr0"
    (ptr (i 0) _ro _t)
    (goal (nonnull (i 0)))
    (addArgs [P.E _ro; P.E _t] pf)
*)

(**** LONGS ****)
let _ = 
  let [_l1; _l2; _h1; _h2] = mkVars 4 in
  simpleAx "packlong"
    (oof (pack2 _l1 _h1) jlong)
    (A.sgIsGround _l1 &&&
     A.sgIsGround _h1 &&&
     A.sgCUT)
    (addArgs [pa2t (P.E _l1); pa2t (P.E _h1)] pf);
  simpleAx "addlong"
    (oof (pack2 (add _l1 _l2) (adc _h1 _h2 (add _l1 _l2))) jlong)
    (A.sgCUT &&&
     goal (oof (pack2 _l1 _h1) jlong) &&&
     goal (oof (pack2 _l2 _h2) jlong))
    pf;
  simpleAx "sublong"
    (oof (pack2 (sub _l1 _l2) (sbb _h1 _h2 (sub _l1 _l2))) jlong)
    (A.sgCUT &&&
     goal (oof (pack2 _l1 _h1) jlong) &&&
     goal (oof (pack2 _l2 _h2) jlong))
    pf
  


(******** JUMP TABLES *********)
let _ = 
  let [_e; _m; _tstart; _idx; _len; _off] = mkVars 6 in
  simpleAx "inJmpTab"
    (injmptable _e _tstart)
    (goal (oof _e (jinjtab _tstart)) &&&
     A.sgCUT)
    pf;
  simpleAx "fetchJmpTable"
    (ptraux _tstart (jtable _len) 
            (add (mul _idx (i 4)) (i 0)) ro (jinjtab _tstart))
    (goal (ult _idx _len) &&&
     A.sgCUT)
    pf;
  simpleAx "fetchJmpTableImm"
    (ptraux _tstart (jtable _len) _off ro (jinjtab _tstart))
    (A.sgCHECK (fun prg s acc -> 
      match fstunroll _off s with
        L.Int off when (L.to_int off) land 3 = 0 -> 
          let off = L.to_int off in
          let idx = off asr 2 in
          let tstart = L.unrollUVarsError s _tstart in
          let res = A.eunify prg (i idx, L.emptySubst) (_idx, s) in
          (res, p2t (P.G("eqCheck", [ P.E (add tstart (mul (i idx) (i 4)));
                                      P.E (add tstart (i off)) ])) :: acc)
      | _ -> (false, acc)) &&&
     goal (ult _idx _len))
    pf;
  simpleAx "baseTypeJmpTab"
    (baseType _tstart (jtable _len))
    (isassume (jmptable _tstart _len) &&&
     A.sgCUT)
    (fun _ _ _ [ pf] -> t2p pf)

(* A special rule that turns out to be used for jump tables *)
let _ =
  let [_e; _len] = mkVars 2 in
  simpleAx "arithjump"
    (ult (sub _e (i 1)) _len)
    (isassume (cle (sub _e _len)) &&&
     isassume (cnlt (sub _e (i 1))) &&&
     A.sgCUT)
    pf
    

(********* EXCEPTIONS ******)
let replacedHandler exc n m n' = L.PFun("replacedHandler", [exc; n; m; n'])
let jexchandler = nullary "jexchandler"

let _ =
  let [_exc; _m; _new] = mkVars 3 in
  simpleAx 
    ~name:"excinst"
    ~conc:(replacedHandler _exc _new (upd 4 _m _exc _new) _new)
    ~sub:(A.sgCUT &&&
          isassume (ptr _exc ro jexchandler) &&&
          A.sgCUT &&&
          goal (oof _m mem))
    ~mkprf:(addArgs [pa2t (P.E _new)] pf)



let entries =
  { S.init     = (fun () -> ());
    S.snapshot = (fun () -> A.debug := if !debug then 1 else 0; A.snapshot theProg);
    S.undo     = (fun () -> A.debug := if !debug then 1 else 0; A.undo theProg);
    S.assume   = assume;
    S.proveIt  = proveIt;
    S.debug    = debug;
    S.ident    = tcjid;
    S.witness  = (fun w -> w);
    S.descr    = "Java typing rules";
  } 
