// annot_aux.cc (renamed from: vcgen_aux.cc)
// I've extended the C++ AST with new classes; this file provides
// definitions for the functions that Elsa otherwise defines for
// the original AST classes

#include "cc.ast.gen.h"        // (extended) C++ AST
#include "cc_env.h"            // Env
#include "cc_print.h"          // PrintEnv
#include "exttype.h"           // EVariable, ext()

// 7/28/04: I've reclaimed the bit that used to be DF_LOGIC since
// I wanted another in Elsa, and didn't appear to be using it here.
#define DF_LOGIC DF_NONE


// ---------------------- TF_macro --------------------------
// cc_tcheck.cc
FakeList<ASTTypeId> *tcheckFakeASTTypeIdList(
  FakeList<ASTTypeId> *list, Env &env, DeclFlags dflags, DeclaratorContext context);

void TF_macro::itcheck(Env &env)
{
  FOREACH_ASTLIST_NC(Macro, macros, iter) {
    iter.data()->tcheck(env);
  }
}


void TF_macro::print(PrintEnv &env)
{
  FOREACH_ASTLIST_NC(Macro, macros, iter) {
    iter.data()->print(env);
  }
}


// -------------------- TF_changeGroup ---------------------
// cc_tcheck.cc
void tcheckPQName(PQName *&name, Env &env, Scope *scope = NULL,
                  LookupFlags lflags = LF_NONE);

// moved this here from c2vml_main.cc
void tcheckE_variableList(Env &env, ASTList<E_variable> &list)
{
  FOREACH_ASTLIST_NC(E_variable, list, iter) {
    E_variable *e = iter.data();
    xassert(!e->ambiguity);        // should never be ambiguous

    tcheckPQName(e->name, env);
    e->var = env.lookupPQVariable(e->name);
    if (!e->var) {
      env.error(e->name->loc, stringc
        << "there is no variable called `" << *(e->name) << "'");
      continue;
    }

    if (e->var->hasFlag(DF_TYPEDEF) &&      // would mean a field changegroup
        !e->var->type->isCompoundType()) {
      env.error(e->name->loc,
        "if a type is mentioned in a 'changes' clause, it must be "
        "a struct/class/union");
    }
  }
}

void TF_changeGroup::itcheck(Env &env)
{
  // check the referenced names, binding them to the corresponding
  // Variables
  tcheckE_variableList(env, group);

  EVariable *v = new EVariable(loc, name->getName(),
                               env.getSimpleType(ST_INT), DF_NONE);
  v->changeGroup = this;
  env.addVariable(v);
}


// exttype.cc
string stringizeE_variableList(ASTList<E_variable> &list);

void TF_changeGroup::print(PrintEnv &env)
{
  env << "  change_group " << name->getName() << " = "
      << stringizeE_variableList(group) << ";\n";
}


// -------------------- TF_axiom ----------------------
void TF_axiom::itcheck(Env &env)
{
  ax->tcheck(env, ax);
}


void TF_axiom::print(PrintEnv &env)
{
  env << "axiom ";
  if (label) {
    env << label << " ";
  }
  env << "{\n";
  ax->print(env);
  env << "}\n";
}


// ---------------- TF_option ----------------
void TF_option::itcheck(Env &env)
{}


void TF_option::print(PrintEnv &env)
{
  env << "verifier_option((" << optionName << "))\n";
}


// ------------------ Macro -------------------
void Macro::tcheck(Env &env)
{
  Scope *s = env.enterScope(SK_FUNCTION /*...*/, "macro");

  // typecheck the bindings; this disambiguates any ambiguous
  // type-ids, and adds them to the environment (I'm not really sure
  // what I should pass for 'isParameter')
  params = tcheckFakeASTTypeIdList(params, env, DF_LOGIC, DC_MACRO);

  body->tcheck(env, body);

  env.exitScope(s);
}


void astTypeIdList_print(FakeList<ASTTypeId> *vars, PrintEnv &env)
{
  env << "(";
  int ct=0;
  FAKELIST_FOREACH(ASTTypeId, vars, iter) {
    if (ct++ > 0) {
      env << ", ";
    }

    env << iter->decl->var->toCStringAsParameter();
  }
  env << ")";
}

void Macro::print(PrintEnv &env)
{
  env << "define " << name.str;
  if (!params) {
    env << " = ";
    body->print(env);
    env << ";\n";
  }
  else {
    astTypeIdList_print(params, env);
    env << " {\n";
    body->print(env);
    env << "}\n\n";
  }
}


// --------------------- S_assert ------------------------
void S_assert::itcheck(Env &env)
{
  // type-check the expression in the ordinary way
  expr->tcheck(env, expr);
}


void S_assert::iprint(PrintEnv &env)
{
  env << "  " << (pure? "pure_" : "") << "assert(";
  expr->print(env);
  env << ");\n";
}


void S_assert::icfg(CFGEnv &env)
{
  // no flow of control within an S_assert
}


// --------------------- S_assume ------------------------
void S_assume::itcheck(Env &env)
{
  // type-check the expression in the ordinary way
  expr->tcheck(env, expr);
}


void S_assume::iprint(PrintEnv &env)
{
  env << "  assume(";
  expr->print(env);
  env << ");\n";
}


void S_assume::icfg(CFGEnv &env)
{
  // no flow of control within an S_assume
}


// --------------------- S_invariant ------------------------
void S_invariant::itcheck(Env &env)
{
  // type-check the expression in the ordinary way
  expr->tcheck(env, expr);
}


void S_invariant::iprint(PrintEnv &env)
{
  env << "  invariant(";
  expr->print(env);
  env << ");\n";
}


void S_invariant::icfg(CFGEnv &env)
{
  // no flow of control within an S_invariant
}


// ----------------------- E_quantifier -----------------
Type *E_quantifier::itcheck_x(Env &env, Expression *&replacement)
{
  Scope *s = env.enterScope(SK_FUNCTION /*...*/, "quantifier");

  // typecheck the bindings; this disambiguates any ambiguous
  // type-ids, and adds them to the environment
  vars = tcheckFakeASTTypeIdList(vars, env,
    DF_LOGIC | (isForall? DF_UNIVERSAL : DF_EXISTENTIAL),
    DC_E_QUANTIFIER);

  if (patterns) {
    patterns->tcheck(env, patterns);
  }

  body->tcheck(env, body);

  env.exitScope(s);

  return body->type;
}


void E_quantifier::iprint(PrintEnv &env)
{
  env << (isForall? "forall " : "exists ");
  if (label) {
    env << label << " ";
  }

  astTypeIdList_print(vars, env);

  if (patterns) {
    env << " ";
    patterns->print(env);
  }

  env << ". ";
  body->print(env);
}


// -------------------------- E_let ---------------------
Type *E_let::itcheck_x(Env &env, Expression *&replacement)
{
  Scope *s = env.enterScope(SK_FUNCTION /*...*/, "let");

  Variable *v = env.tfac.makeVariable(
    name.loc,
    name.str,
    env.getSimpleType(ST_INT),    // will do for now
    DF_LOGIC,
    NULL /*tunit*/);
  env.addVariable(v);    // can't fail, just opened a new scope

  body->tcheck(env, body);

  env.exitScope(s);

  return body->type;
}


void E_let::iprint(PrintEnv &env)
{
  env << tovc();
}


// -------------------------- E_pre ---------------------
Type *E_pre::itcheck_x(Env &env, Expression *&replacement)
{
  expr->tcheck(env, expr);
  return expr->type;
}


void E_pre::iprint(PrintEnv &env)
{
  env << "pre(";
  expr->print(env);
  env << ")";
}


// -------------------------- E_vc_expression ---------------------
Type *E_vc_expression::itcheck_x(Env &env, Expression *&replacement)
{
  expr->tcheck(env, expr);
  return expr->type;
}


void E_vc_expression::iprint(PrintEnv &env)
{
  env << "vc_expression(";
  expr->print(env);
  env << ")";
}


// ------------------------- Declarator -----------------------
// this replaces Elsa's mid_tcheck
void Declarator::mid_tcheck(Env &env, Tcheck &dt)
{
  // call into the original one
  orig_mid_tcheck(env, dt);
  if (!var) {
    return;
  }

  // now search for D_varattributes in the IDeclarator
  IDeclarator *idecl = decl;
  while (idecl) {
    if (idecl->isD_varattribute()) {
      StringRef attr = idecl->asD_varattribute()->attr;
      if (!ext(var)->attachAttribute(attr)) {
        env.error(stringc << "unknown variable attribute: " << attr);
      }
    }

    idecl = idecl->getBase();
  }
}


// ------------------------- D_varattribute ---------------
void D_varattribute::tcheck(Env &env, Declarator::Tcheck &dt)
{
  base->tcheck(env, dt);

  #if 0    // old
  // attach the attribute
  if (!ext(dt.var)->attachAttribute(attr)) {
    env.error(stringc << "unknown variable attribute: " << attr);
  }
  #endif // 0
}


PQName const *D_varattribute::getDeclaratorIdC() const
{
  return base->getDeclaratorId();
}


IDeclarator const *D_varattribute::getBaseC() const
{
  return base;
}


// EOF
