// c2v_lexer.cc            see license.txt for copyright and terms of use
// code for c2v_lexer.h

#include "c2v_lexer.h"        // C2VLexer

char const *toString(C2VToken type)
{
  // corresponds to list in c2v_ext_tokens.h
  static char const * const map[] = {
    "emit",
    "for_each_field",
    "of",
    "return",

    "matchExpr",
    "matchStmt",
    "matchType",

    "<embedded_code>",
    
    "<variable_name>",

    "<patv_identifier>",

    "<cpatv_statement>",
    "<cpatv_expression>",
    "<cpatv_type>",
    "<cpatv_variable>",
    "<cpatv_field>",

    "<vpatv_variable>",
    "<vpatv_expression>",
    "<vpatv_predicate>",
    "<vpatv_statement>",
    "<vpatv_field>",
  };
  ASSERT_TABLESIZE(map, NUM_C2V_TOKENS - NUM_VML_TOKENS);
  xassert((unsigned)type < NUM_C2V_TOKENS);

  if ((unsigned)type < NUM_VML_TOKENS) {
    return toString((VMLToken)type);
  }
  else {
    return map[type - NUM_VML_TOKENS];
  }
}


C2VLexer::C2VLexer(StringTable &strtable, char const *fname)
  : VMLLexer(strtable, fname),
    recentMatchKeyword(false),
    embedStartLoc(SL_UNKNOWN),
    embedStart(0),
    embedFinish(0),
    embedded(new CCSubstrate())
{}

C2VLexer::~C2VLexer()
{
  delete embedded;
}


string C2VLexer::tokenDesc() const
{
  switch (type) {
    case C2VTOK_STRING_LIT:
      return stringc << "str(" << (StringRef)sval << ")";

    case C2VTOK_EMBEDDED_CODE:
      return stringc << "embed(" << (StringRef)sval << ")";

    case TOK_INT_LIT:
    case TOK_IDENTIFIER:
      return VMLLexer::tokenDesc();

    default:
      return toString((C2VToken)type);
  }
}

string C2VLexer::tokenKindDesc(int kind) const
{
  return toString((C2VToken)kind);
}


// ------------------ test code -----------------
#ifdef TEST_C2V_LEXER

#include "exc.h"        // xBase
#include "srcloc.h"     // SourceLocManager

int main(int argc, char **argv)
{
  if (argc < 2) {
    cout << "usage: " << argv[0] << " input.c2v\n";
    return 0;
  }

  try {
    SourceLocManager mgr;
    StringTable table;
    C2VLexer lexer(table, argv[1]);
    lexer.tokenFunc(&lexer);

    while (lexer.type != TOK_EOF) {
      cout << toString(lexer.loc) << ": " << lexer.tokenDesc() << "\n";
      lexer.getTokenFunc()(&lexer);
    }

    return lexer.errors? 4 : 0;
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
}

#endif // TEST_C2V_LEXER
