// c2v_parse.cc
// code for c2v_parse.h

#include "c2v_parse.h"      // this module
#include "c2v_parse_c.h"    // c2v_parse_c
#include "c2v_lexer.h"      // C2VLexer
#include "c2v.gr.gen.h"     // C2VParse
#include "owner.h"          // Owner
#include "glr.h"            // GLR
#include "exc.h"            // xbase
#include "cc_lang.h"        // CCLang


C2VFile *parse_C2V_file(StringTable &strTable, CCLang &lang, char const *fname)
{
  // lexer
  C2VLexer lexer(strTable, fname);
  lexer.tokenFunc(&lexer);

  // parser context
  C2VParse ctx(strTable);

  // parse tables (automatically deallocated)
  Owner<ParseTables> tables(ctx.makeTables());

  // parser
  GLR glr(&ctx, tables);

  // parse the input
  SemanticValue treeTop;
  if (!glr.glrParse(lexer, treeTop)) {
    xbase(stringc << "parse error in file " << fname);
  }
  C2VFile *file = (C2VFile*)treeTop;

  // visit the resulting AST, and parse the embedded C fragments
  c2v_parse_c(file, strTable, lang);

  return file;
}


// -------------------- test code --------------------
#ifdef TEST_C2V_PARSE
              
#include <iostream.h>      // cout
#include "strtable.h"      // StringTable
#include "srcloc.h"        // SourceLocManager
#include "trace.h"         // TRACE_ARGS
#include "mysig.h"         // pringSegfaultAddrs

int doit(int argc, char *argv[])
{
  char const *progname = argv[0];
  TRACE_ARGS();

  if (argc <= 1) {
    cout << "usage: " << progname << " [-tr flags] input.c2v\n";
    return 0;
  }

  printSegfaultAddrs();

  CCLang cclang;
  cclang.ANSI_C89();

  SourceLocManager mgr;
  StringTable str;
  C2VFile *file = parse_C2V_file(str, cclang, argv[1]);

  file->debugPrint(cout, 0);
  return 0;
}


int main(int argc, char *argv[])
{
  try {
    return doit(argc, argv);
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
}


#endif // TEST_C2V_PARSE
