// c2v_subst.cc
// substituting VML AST fragments

#include "c2vml.h"          // contains declarations for this module


// ----------------- SubstMVisitor -------------------
class SubstMVisitor : public VMLMVisitor {
public:
  C2VML &env;

public:
  SubstMVisitor(C2VML &e) : env(e) {}

  virtual bool visitVExpr(VExpr *&obj);
  virtual bool visitIdentifier(Identifier *&obj);
};


bool SubstMVisitor::visitVExpr(VExpr *&obj)
{
  if (obj->isVE_patvar()) {
    Identifier *id = obj->asVE_patvar()->id;
    C2VValue *v = env.lookupf(id->name);

    obj = v->getVExpr()->clone();
  }

  return true;
}


bool SubstMVisitor::visitIdentifier(Identifier *&obj)
{
  if (obj->name[0] == '$') {
    obj->name = env.lookupf(obj->name)->getIdentifier();
  }
  
  return true;
}


// ------------------ C2VML subst methods --------------------
VStatement * owner C2VML::substVStatement(VStatement const *pat)
{
  // clone the pattern
  VStatement *ret = pat->clone();

  // run a mutating visitor to replace references to pattern
  // variables with the associated AST fragments
  SubstMVisitor vis(*this);
  vis.mtraverse(ret);

  return ret;
}


VExpr * owner C2VML::substVExpr(VExpr const *pat)
{
  // same as above
  VExpr *ret = pat->clone();
  SubstMVisitor vis(*this);
  vis.mtraverse(ret);
  return ret;
}


// EOF
