// cfgpaths.h            see license.txt for copyright and terms of use
// routines for enumerating CFG paths in the verifier AST

// "Path" for this module is an acyclic sequence of Statements,
// starting at the function-entry point or an invariant, and ending at
// an invariant or a function-exit point.  For statements that have
// multiple execution paths (e.g. for short-circuit evaluation), it
// encodes which execution path to use.

#ifndef CFGPATHS_H
#define CFGPATHS_H

#include "cc.ast.gen.h"     // C++ AST elements
#include "sobjlist.h"       // SObjList

// instrument the AST of a function to enable printing (among other
// things) of paths; returns total # of paths in the function (in
// addition to storing that info in the AST)
int countPaths(Function *func);

// print all paths in this function
void printPaths(Function const *func);
void printPathFrom(SObjList<Statement /*const*/> &path, int index,
                   Statement const *node, bool isContinue);

// count/print for statements
int countExprPaths(Statement const *stmt, bool isContinue);
void printExprPath(int index, Statement const *stmt, bool isContinue);

// count/print for inititializers
int countExprPaths(Initializer const *init);
void printExprPath(int index, Initializer const *init);

// count/print for expressions
int countPaths(Expression *ths);
void printPath(int index, Expression const *ths);

// similar for FullExpression
inline int countPaths(FullExpression *ths)
  { return countPaths(ths->expr); }
inline void printPath(int index, FullExpression const *ths)
  { printPath(index, ths->expr); }

// starting from a point above 'stmt', how many paths go into 'stmt'
// (and possibly beyond)?
int numPathsThrough(Statement const *stmt);

#endif // CFGPATHS_H
