// array2.c
// scull-like array of arrays

// --------- verifier preamble ----------
#include "ptr.model"     // sel/upd/ptr
#include "state.model"   // metaVar
#include "div.model"     // div/mod

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);


// ------------ code to verify ------------
// these are never changed, though not declared 'const' in scull
//
// 9/20/04: I changed these to be 40 and 10 instead of 4000 and 1000
// because the larger numbers cause an arithmetic overflow in
// Simplify during a Simplex pivot operation
int const quantum = 40;     // invariant: quantum > 0
int const qset = 10;        // invariant: qset > 0

// when not NULL
//   data is an owner pointer to qset elements
//   each element is either NULL, or a pointer to quantum non-dead chars
char * /*owner*/ * /*owner*/ data = NULL;


// ------ invariants ------
// invariant of the non-NULL pointers stored in the array
#define okArrayPtr(p) (            \
  p < objct &&                     \
  objsize[p] == quantum            \
)


// invariant of the array at 'data', considering elements
// starting at 'low'
#define okArray(data, low) (                                              \
  /* all the pointers are ok */                                           \
  (forall(int i) PATS(mem[data,i], meta[data,i]).                         \
    low<=i<qset                                                           \
  ==>                                                                     \
    meta[data,i] == live &&   /* pointer is live */                       \
    (mem[data,i]==NULL || okArrayPtr((mem[data,i]))) &&                   \
    mem[data,i] != data       /* none point back into the main array */   \
  )&&                                                                     \
                                                                          \
  /* they're all different from each other */                             \
  (forall(int i, int j) PATS(MPAT(mem[data,i], mem[data,j])).             \
    low<=i<qset &&                                                        \
    low<=j<qset &&                                                        \
    i != j &&                                                             \
    mem[data,i] != NULL &&                                                \
    mem[data,j] != NULL                                                   \
  ==>                                                                     \
    mem[data,i] != mem[data,j]                                            \
  )                                                                       \
)


// global invariant
int global_inv() { return vc_expression(all_of {
  quantum > 0;
  qset > 0;
  objct > NULL;
  data < objct;
  data!=NULL ==> (
    0 < data &&
    //metaVar(data) == live &&
    objsize[data] == qset &&
    okArray(data, 0)
  );
});}


void write(unsigned f_pos, char val)
  changes(mem, meta, tag, objsize, objct, data)
{
  int itemsize = quantum * qset;
  int rest = f_pos % itemsize;

  assert(0 <= rest < itemsize);

  int s_pos = rest / quantum;
  int q_pos = rest % quantum;

  assert(0 <= s_pos < qset);
  assert(0 <= q_pos < quantum);

  if (!data) {
    data = (char * /*owner*/ * /*owner*/)malloc(qset * sizeof(char * /*owner*/));
    // note: assume malloc can't fail

    // data is an owner pointer to qset dead pointers

    memset(data, 0, qset * sizeof(char * /*owner*/));

    // data is an owner pointer to qset NULL pointers
  }

  invariant {
    0 <= rest < itemsize;
    0 <= s_pos < qset;
    0 <= q_pos < quantum;
    f_pos >= 0;
    data!=NULL;
  }

  if (!data[s_pos]) {
    data[s_pos] = (char * /*owner*/)malloc(quantum * sizeof(char));

    // data[s_pos] is an owner pointer to quantum dead chars

    memset(data[s_pos], 0, quantum * sizeof(char));

    // data[s_pos] is an owner pointer to quantum non-dead chars
  }

  invariant {
    0 <= rest < itemsize;
    0 <= s_pos < qset;
    0 <= q_pos < quantum;
    f_pos >= 0;
    data!=NULL;
    mem[data,s_pos]!=NULL;
  }

  data[s_pos][q_pos] = val;
}


// post: data is NULL
void trim()
  changes(mem, meta, tag, data)
{
  if (!data) return;

  for (int ii=0; ii<qset; ii++) {
    // for 0<=j<ii, data[j] is dead
    standalone_invariant {
      // global_inv stuff...
      quantum > 0;
      qset > 0;
      objct > NULL;
      0 < data < objct;
      //metaVar(data) == live;
      objsize[data] == qset;
  
      0 <= ii < qset;
      forall(int j) PATS(meta[data,j], mem[data,j]).
        0<=j<ii ==> (meta[data,j]==dead ||  // low half: dead or NULL
                     mem[data,j]==NULL);
      okArray(data, ii);                                 // high half: ok
    }

    if (data[ii]) {
      free(data[ii]);
    }
  }
  
  free(data);
  data = NULL;
}


/*
  Errors to catch:
    - failure to call memset(0) in either place in write()
    - failure to free all the data[i]'s in trim()
*/
