// bubblesort.c
// based on pale/Examples/bubblesort_simple.pale

// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar


// --------------- code to verify -----------------
typedef struct Node {
  int value;
  struct Node *next;    // invariant: NULL or valid
  struct Node *prev;    // ghost: thing whose 'next' points at me
} Node;


typedef int bool;
#define false 0
#define true 1


// global invariants
int global_inv() { return vc_expression(all_of {
  tag[NULL]==0;                         // zero tag at NULL, i.e. not an object

  // all Nodes
  
  // declared type of 'next' is accurate
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==> tag[n->next] == Node_tag;

  // injectivity of 'next'
  forall(Node *n) PATS(n->next).
    tag[n]==Node_tag && n->next!=NULL ==> n->next->prev == n;

  // declared type of 'prev' is accurate
  forall(Node *n) PATS(n->prev).
    tag[n]==Node_tag && n->prev!=NULL ==> tag[n->prev]==Node_tag;

  // reverse injectivity... needed b/c I added the assertion about
  // prev's type, which itself is needed for me to be able to
  // deref the prev ptr..
  forall(Node *n) PATS(n->prev).
    tag[n]==Node_tag && n->prev!=NULL ==> n->prev->next == n;
});}


Node *bubblesort(Node *x)
  pre(x!=NULL ==> tag[x]==Node_tag && x->prev==NULL)
  post(result!=NULL ==> tag[result]==Node_tag && result->prev==NULL)
  changes(Node)
{
  Node *p, *y, *yn, *t;
  bool change;
  change = true;

  // this outer loop does the bubble process over and over until
  // quiescence; 'x' always points to the beginning of the list
  while (change) {
    invariant all_of {
      x!=NULL ==> tag[x]==Node_tag && x->prev==NULL;
    };

    p = NULL;
    change = false;
    y = x;
    if (x != NULL) {
      yn = y->next;
    }
    else { /*BUG FIX!*/
      yn = NULL;
    }
    
    // the inner loop moves 'y' down the list; 'yn' is kept equal
    // to 'y->next'; each iteration considers swapping 'y' and 'yn'
    // in the list; 'p' stays one step beind 'y'; just before a swap
    // is done, 't' is set to 'yn->next'
    //
    // so the typical picture (just before a swap) is
    //
    //       x          p     y     yn    t
    //       |          |     |     |     |
    //       V          V     V     V     V
    //      +--+       +--+  +--+  +--+  +--+       +--+
    //      |  |->...->|  |->|  |->|  |->|  |->...->|  |->NULL
    //      +--+       +--+  +--+  +--+  +--+       +--+
    //
    while (yn!=NULL) {
      invariant all_of {
        tag[x]==Node_tag;
        x->prev==NULL;
        tag[y]==Node_tag;          // y!=NULL;
        tag[yn]==Node_tag;
        yn==y->next;
        y!=x ==> tag[p]==Node_tag && p->next==y && p!=y && p!=yn;
        y==x ==> p==NULL;
        //x<next*>y;               // cannot say
        //x==NULL ==> yn==NULL;    // not necessary

        // There is a subtlety with regard to disequality here: it is
        // of course intended that p!=y!=yn, but because I have not
        // included acyclicity invariants that must be explicitly
        // stated.  However, injectivity *is* strong enough to carry
        // this ternary disequality throughout the data strucure.
        // (The disequality constraints on 'p' are above, in the case
        // when it is not NULL.)
        y != yn;
      };

      if (y->value > yn->value) {
        t = yn->next;
        change = true;
        y->next = t;
        if (t != NULL) {
          t->prev = y;                 // ghost update
        }
        yn->next = y;
        y->prev = yn;                  // ghost update
        if (p==NULL) {
          x = yn;
          yn->prev = NULL;             // ghost update
        }
        else {
          p->next = yn;
          yn->prev = p;                // ghost update
          assert(x != y);
          assert(yn->next == y);
        }
        p = yn;
        yn = t;
      }
      else {
        p = y;
        y = yn;
        yn = y->next;
      }
    }
  }
  return x;
}
