// funcall.c
// experimenting with function calls, pre/post, etc.


// ---------------- verifier preamble --------------
// more or less true globals
//int mem;
//int meta;
//int objct;

int fresh_value();


// -------------- code to verify -------------
// more globals
int x;
int y;


int global_inv() { return vc_expression(all_of {
  1==1;       // "true"
});}


// modifies nothing, no pre/post
void noop();


// modifies nothing, but reliably returns 3
int three()
  post(result == 3)
{
  return 3;
}


// increments x
void incx()
  post(x == pre(x)+1)     // pre(x) refers to x in the pre-state
  changes(x)
{
  x = x+1;
}


// increments x by some positive amount
void addx(int amt)
  pre(amt > 0)
  post(x > pre(x))        // deliberate underspecification
  changes(x)
{
  x = x + amt;
}


void main()
  changes(x,y)
{
  x = three();
  y = 4;

  assert(x==3 && y==4);

  noop();

  assert(x==3 && y==4);

  incx();

  assert(x==4 && y==4);

  addx(2);

  assert(x>4 && y==4);

  addx(x);

  assert(x>4 && y==4);
}


void bad()
{
  //ERROR(1): x = 5;    // not declared in 'writes'
}

void good()
{
  int orig_x = x;
  
  x = 7;
  
  x = orig_x;
}

void good2()
{
  int orig_x = x;

  while (fresh_value()) {
    invariant(orig_x == pre(x));    // infer...
    x++;
  }

  x = orig_x;
}

void good3()
{
  int orig_x = x;

  while (fresh_value()) {
    invariant(orig_x == pre(x));    // infer...
    x++;
  }

  while (fresh_value()) {
    invariant(orig_x == pre(x));    // infer...
    x++;
  }

  x = orig_x;
}

void good4()
  pre(x == 5)    // so as long as x==5 at end, it has not changed
{
  while (fresh_value()) {
    invariant(1==1);
    x++;             
  }
  
  x = 5;
}

void bad2()
{
  int y=1;
  while (1==1) {
    int x;
    if (x==0) break;
    y++;
    invariant (x!=0);
  }

  // this *is* provable if I do not do the modct increment
  // at the declaration of 'x'  
  //ERROR(2): assert(y!=1 ==> 1==0);
}





// EOF
