// invar.c
// data structure invariants of a structure with embedded array


// --------- verifier preamble ----------
#include "ptr.model"        // sel/upd/ptr
#include "state.model"      // metaVar

//asm("verifier: include \"div.vc\";");

// C functions that are modeled directly
void *malloc(int size);
void free(void *ptr);
void memset(void *ptr, int val, int length);


// ------------------ code to verify -----------------------
// this structure's "invariants" contribute to the notion
// of validity for instances of the struct
typedef struct Foo {
  //int use_flat_struct_model;        // to support the array...
  int x;        // invariant A: 2 <= x <= 5
  int arr[6];   // invariant B: arr[i] >= 3
                // invariant C: arr[x] == 24
} Foo;

// the object at 'broken' does not necessarily respect
// any of invariants A,B,C
Foo *broken = NULL;


int global_inv() { return vc_expression(all_of {
  0 < objct;
  tag[NULL]==0;

  forall(Foo *addr) PATS(tag[addr]).
    tag[addr]==Foo_tag ==> 0 < addr < objct;

  // invariant A
  forall(Foo *addr) PATS(addr->x).
    tag[addr]==Foo_tag && addr!=broken ==>
      2 <= addr->x <= 5;

  // invariant B
  forall(Foo *addr, int i) PATS(addr->arr[i]).
    tag[addr]==Foo_tag && addr!=broken && 0<=i<6 ==>
      addr->arr[i] >= 3;

  // invariant C
  forall(Foo *addr) PATS(addr->arr[addr->x]).
    tag[addr]==Foo_tag && addr!=broken ==>
      addr->arr[addr->x] == 24;

});}


// TODO: let 'anotherFunc' appear below 'func', with just
// a prototype here


// pre/post: p is valid
void anotherFunc(Foo * /*serf*/ p)
  pre(broken==NULL && tag[p] == Foo_tag)     // 'p' is a valid Foo
  changes(Foo)
{
  assert(p->arr[p->x] == 24);

  if (p->arr[p->x] != 24) {
    assert(0==1);    // unreachable
  }

  p->x = 5;

  // p is not necessarily valid now

  p->arr[5] = 24;

  // p is valid again
  
  //ERROR(1): p->x = 4;
}


void func()
  pre(broken==NULL)
  post(broken==NULL)
  changes(Foo, tag, objct, objsize)
{
  Foo * /*owner*/ p = (Foo * /*owner*/)malloc(sizeof(Foo));
  broken = p;

  p->x = 2;

  // invariant A of *p is now satisfied

  for (int i=0; i<6; i++) {
    // invariant A of *p is satisfied

    invariant {
      // components of Foo's invariant that have been established
      2 <= p->x <= 5;       // infer?

      // invariants of this loop
      0 <= i;               // infer?
      forall(int j) PATS(p->arr[j]).
        0<=j<i ==> p->arr[j] >= 3;
    }

    p->arr[i] = i + 3;
  }

  // invariants A and B of *p are now satisfied

  p->arr[p->x] = 24;

  // p is now valid:
  //   p is an owner pointer to Foo-sized memory
  //   all invariants of *p are satisfied
  broken = NULL;

  anotherFunc(p);

  // p is still valid

  free(p);
}

