; memory.thy      -*- Lisp -*-
; theory of memory operations, e.g. sel and upd

(BG_PUSH (AND

  ; --------- sel/upd -----------

  ; (sel obj index): select subobject 'index' from 'obj'; the index
  ;                  could be an array index, or a field number
  ; (upd obj index val): yield a modified 'obj' where (sel obj index)
  ;                      equals 'val' in the yielded object
  ; (initmem val): all values are 'val'

  ; reading from the most recently-written field
  (FORALL (obj index val)
    (EQ
      (sel (upd obj index val) index)          ; obj[index := val].index
      val
    ))

  ; reading from a different field
  (FORALL (obj index1 index2 val)
    (IMPLIES
      (NEQ index1 index2)
      (EQ
        (sel (upd obj index1 val) index2)      ; obj[index1 := val].index2
        (sel obj index2)                       ; obj.index2
      )
    ))

  ; reading from an initmem
  (FORALL (index val)
    (EQ
      (sel (initmem val) index)
      val
    ))

))
