// prover.h
// interface to a theorem prover
//
// One might imagine abstracting this interface further, to achieve
// greater independence from Simplify itself, but I have not done
// so.  Among other things, toSexp (etc.) would also need to be
// abstracted.  So, right now, it's basically just a thin wrapper
// around the IPC pipes.

#ifndef PROVER_H
#define PROVER_H

#include "str.h"       // string

// try to prove predicate represented as an sexp 'str',
// and report success or failure
bool runProver(rostring str);


// more refined version of the above, as an object I can control
class Prover {
private:     // data
  int toSimplify, fromSimplify;    // pipe file descriptors
  int predicateLog;                // file descriptor of log of calls

private:     // disallowed
  Prover(Prover&);
  void operator= (Prover&);

public:      // funcs
  Prover();                        // start child process
  ~Prover();                       // terminate process

  // begin logging predicates to a given file
  void openLog(rostring fname);

  // read a file, and feed all of its contents to Simplify; do
  // not expect any output from Simplify
  void feedFile(rostring fname);

  // feed a string to Simplify, again not expecting output
  void feed(rostring str);

  // prove a predicate; returns true if the predicate is proven,
  // false if it is not; 'str' should have a newline at the end;
  // if 'counterexample' is not NULL, and Simplify yields a
  // counterexample, put it in that string
  bool prove(rostring str, string *counterexample = NULL);
};

#endif // PROVER_H
