; ptr.thy
; theory of pointers as objects with which to index memory;
; this is the second major model of pointers I've encoded
; in a .thy file

; requires: memory.thy

(BG_PUSH (AND

  ; (ptr addr field): pointer to object at 'addr', field 'field';
  ; this can directly index into memory, or a meta-memory
  
  ; key property:
  ;   (ptr a1 f1) = (ptr a2 f2)  <==>  a1=a2 && f1=f2
  ; <=: congruence
  ; =>: injectivity:
  (FORALL (a1 f1 a2 f2)
    (IMPLIES
      (EQ (ptr a1 f1) (ptr a2 f2))
      (AND (EQ a1 a2) (EQ f1 f2))
    ))

  ; (ptr_addr ptr): select addr component
  (FORALL (a f)
    (EQ
      (ptr_addr (ptr a f))
      a
    ))

  ; (ptr_field ptr): select field component
  (FORALL (a f)
    (EQ
      (ptr_field (ptr a f))
      f
    ))

  ; (updslice M a v): update all slots of an object
  (FORALL (mem addr val p)
    (IMPLIES
      (EQ addr (ptr_addr p))      ; if pointer is to same object
      (EQ
        (sel (updslice mem addr val) p)     ; then reading from it
        val                                 ; yields object-wide written value
      )
    ))
  (FORALL (mem addr val p)
    (IMPLIES
      (NEQ addr (ptr_addr p))     ; if pointer is to different object
      (EQ
        (sel (updslice mem addr val) p)     ; then reading from it
        (sel mem p)                         ; yields same as before the write
      )
    ))

  ; (updrange M a f_low f_high v): update fields f_low through f_high-1 with v
  (FORALL (mem addr f_low f_high val p)
    (IMPLIES
      (AND                        ; if pointer to same object, in field range
        (EQ addr (ptr_addr p))
        (>= (ptr_field p) f_low)
        (< (ptr_field p) f_high)
      )
      (EQ                         ; reading yields val
        (sel (updrange mem addr f_low f_high val) p)
        val
      )
    ))
  (FORALL (mem addr f_low f_high val p)
    (IMPLIES
      (NOT (AND                   ; if pointer to diff't object, or out of range
        (EQ addr (ptr_addr p))
        (>= (ptr_field p) f_low)
        (< (ptr_field p) f_high)
      ))
      (EQ                         ; reading skips this record
        (sel (updrange mem addr f_low f_high val) p)
        (sel mem p)
      )
    ))



))

