// react.c
// reactive system state transitions, with state-dependent invariant


// --------- verifier preamble ----------


// ------------------ code to verify -------------------
typedef enum State { S_UNLOADED, S_CLOSED, S_OPEN } State;

State state = S_UNLOADED;     // ghost

// invariant:
//   state==S_UNLOADED ==> x==0
//   state==S_CLOSED   ==> x==2
//   state==S_OPEN     ==> x==5
int x = 0;


// global invariants
int global_inv() { return vc_expression(all_of {
  (state==S_UNLOADED && x==0 ||
   state==S_CLOSED   && x==2 ||
   state==S_OPEN     && x==5);
});}


int rand()
  changes()
;


// pre: state==S_UNLOADED
// post: result==0 && state==S_CLOSED ||
//       result<0  && state==S_UNLOADED
int init()
  pre(state==S_UNLOADED)
  post(result==0 && state==S_CLOSED ||
       result<0  && state==S_UNLOADED)
  changes(state, x)
{
  if (rand()) {
    state = S_CLOSED;         // ghost
    x = 2;
    return 0;
  }
  else {
    return -1;
  }
}


// pre: state==S_CLOSED
// post: state==S_UNLOADED
void cleanup()
  pre(state==S_CLOSED)
  post(state==S_UNLOADED)
  changes(state, x)
{
  state = S_UNLOADED;         // ghost
  x = 0;
}


// pre: state==S_CLOSED
// post: result==0 && state==S_OPEN ||
//       result<0  && state==S_CLOSED
int open()
  pre(state==S_CLOSED)
  post(result==0 && state==S_OPEN ||
       result<0  && state==S_CLOSED)
  changes(state, x)
{
  if (rand()) {
    state = S_OPEN;           // ghost
    x = 5;
    return 0;
  }
  else {
    return -1;
  }
}


// pre: state==S_OPEN
// post: state==S_CLOSED
void close()
  pre(state==S_OPEN)
  post(state==S_CLOSED)
  changes(state, x)
{
  state = S_CLOSED;           // ghost
  x = 2;
}


// pre: state==S_OPEN
// post: state==S_OPEN
void read()
  pre(state==S_OPEN)
  post(state==S_OPEN)
{
  assert(x == 5);
}


/*
  Errors to catch:
    - returning +1 when -1 was intended
    - failure to set x properly
*/
