// scull.c
// source code to 'scull' driver, all in one file


// --------- verifier preamble ----------
// for now, use asm hack to put the includes in
asm("verifier: include \"ptr.vc\";");
asm("verifier: include \"state.vc\";");
asm("verifier: include \"div.vc\";");
asm("verifier: include \"bitwise.vc\";");    // bitand
asm("verifier: include \"reach.vc\";");      // canReach, noCycles
asm("verifier: include \"disjoint.vc\";");   // disjoint
asm("verifier: include \"objct.vc\";");      // reachAllocated
asm("verifier: include \"cycles.vc\";");     // noCycles

// C functions that are modeled directly
void *malloc(int size);

void memset(void *ptr, int val, int length);

// declarations of model functions
int sel(int mem, int p);
int ptr(int addr, int field);
int upd(int mem, int addr, int val);
int updflatrange(int mem, int lo, int hi, int v);
int updbothranges(int mem, int loAddr, int hiAddr, int loIndex, int hiIndex, int v);
int canReach(int mem, int field, int src, int dest);
int canReachWO(int mem, int field, int src, int dest, int avoid);
int noCycles(int mem, int field, int ptr);
int disjoint(int mem, int field, int p, int q);
int ptr_field(int ptr);
int ptr_addr(int ptr);
int metaVar(int whatever);

// need to somehow say this is a constant symbolic value, not a global..
// let's try 'const'
int const dead;
int const live;
//int const state;

// more or less true globals
int mem;      // for arrays
int meta;     // for arrays
int tag;
int objsize;
int objct;
             

// experiment; problem is this wants the global invariant to
// hold, but I just want to model it simply.. modified vcgen.cc
// to do so, so this pre/post isn't used now
void kfree(void *freedPtr)
  pre(tag[freedPtr] != 0)
  post(tag == pre(tag){freedPtr := 0})
  changes(tag)
;



// ------------------ from the kernel headers -----------------
// unknown source
typedef int loff_t;
typedef int ssize_t;
typedef int size_t;
typedef int filldir_t;
typedef int kdev_t;
typedef int pid_t;


// linux/kernel.h
#define printk printf
void printf(char const *fmt, ...);
#define	KERN_EMERG	"<0>"	/* system is unusable			*/
#define	KERN_ALERT	"<1>"	/* action must be taken immediately	*/
#define	KERN_CRIT	"<2>"	/* critical conditions			*/
#define	KERN_ERR	"<3>"	/* error conditions			*/
#define	KERN_WARNING	"<4>"	/* warning conditions			*/
#define	KERN_NOTICE	"<5>"	/* normal but significant condition	*/
#define	KERN_INFO	"<6>"	/* informational			*/
#define	KERN_DEBUG	"<7>"	/* debug-level messages			*/


// linux/fs.h
/*
 * NOTE:
 * read, write, poll, fsync, readv, writev can be called
 *   without the big kernel lock held in all filesystems.
 */
typedef struct file_operations {
//	struct module *owner;
	loff_t (*llseek) (struct file *, loff_t, int);
	ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	int (*readdir) (struct file *, void *, filldir_t);
	unsigned int (*poll) (struct file *, struct poll_table_struct *);
	int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
	int (*mmap) (struct file *, struct vm_area_struct *);
	int (*open) (struct inode *, struct file *);
	int (*flush) (struct file *);
	int (*release) (struct inode *, struct file *);
//  	int (*fsync) (struct file *, struct dentry *, int datasync);
//  	int (*fasync) (int, struct file *, int);
//  	int (*lock) (struct file *, int, struct file_lock *);
//  	ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
//  	ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
//  	ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
//  	unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
} file_operations;


typedef int ModuleId;
enum FileState { FS_CLOSED, FS_OPEN };


// linux/fs.h
typedef struct file {
//  	struct list_head	f_list;
//  	struct dentry		*f_dentry;
//  	struct vfsmount         *f_vfsmnt;
	struct file_operations	*f_op;
//  	atomic_t		f_count;
  	unsigned int 		f_flags;
//  	mode_t			f_mode;
  	loff_t			f_pos;
//  	unsigned long 		f_reada, f_ramax, f_raend, f_ralen, f_rawin;
//  	struct fown_struct	f_owner;
//  	unsigned int		f_uid, f_gid;
//  	int			f_error;

//  	unsigned long		f_version;

//  	/* needed for tty driver, and maybe others */
  	void			*private_data;

//  	/* preallocated helper kiobuf to speedup O_DIRECT */
//  	struct kiobuf		*f_iobuf;
//  	long			f_iobuf_lock;

        // annotation (ghost) variables
        ModuleId                id;
        enum FileState               state;
} file;


// linux/fs.h
typedef struct inode {
//  	struct list_head	i_hash;
//  	struct list_head	i_list;
//  	struct list_head	i_dentry;

//  	struct list_head	i_dirty_buffers;
//  	struct list_head	i_dirty_data_buffers;

//  	unsigned long		i_ino;
//  	atomic_t		i_count;
//  	kdev_t			i_dev;
//  	umode_t			i_mode;
//  	nlink_t			i_nlink;
//  	uid_t			i_uid;
//  	gid_t			i_gid;
  	kdev_t			i_rdev;
//  	loff_t			i_size;
//  	time_t			i_atime;
//  	time_t			i_mtime;
//  	time_t			i_ctime;
//  	unsigned int		i_blkbits;
//  	unsigned long		i_blksize;
//  	unsigned long		i_blocks;
//  	unsigned long		i_version;
//  	struct semaphore	i_sem;
//  	struct semaphore	i_zombie;
//  	struct inode_operations	*i_op;
//  	struct file_operations	*i_fop;	/* former ->i_op->default_file_ops */
//  	struct super_block	*i_sb;
//  	wait_queue_head_t	i_wait;
//  	struct file_lock	*i_flock;
//  	struct address_space	*i_mapping;
//  	struct address_space	i_data;
//  	struct dquot		*i_dquot[MAXQUOTAS];
//  	/* These three should probably be a union */
//  	struct list_head	i_devices;
//  	struct pipe_inode_info	*i_pipe;
//  	struct block_device	*i_bdev;
//  	struct char_device	*i_cdev;

//  	unsigned long		i_dnotify_mask; /* Directory notify events */
//  	struct dnotify_struct	*i_dnotify; /* for directory notifications */

//  	unsigned long		i_state;

//  	unsigned int		i_flags;
//  	unsigned char		i_sock;

//  	atomic_t		i_writecount;
//  	unsigned int		i_attr_flags;
//  	__u32			i_generation;
//  	union {
//  		struct minix_inode_info		minix_i;
//  		struct ext2_inode_info		ext2_i;
//  		struct ext3_inode_info		ext3_i;
//  		struct hpfs_inode_info		hpfs_i;
//  		struct ntfs_inode_info		ntfs_i;
//  		struct msdos_inode_info		msdos_i;
//  		struct umsdos_inode_info	umsdos_i;
//  		struct iso_inode_info		isofs_i;
//  		struct nfs_inode_info		nfs_i;
//  		struct sysv_inode_info		sysv_i;
//  		struct affs_inode_info		affs_i;
//  		struct ufs_inode_info		ufs_i;
//  		struct efs_inode_info		efs_i;
//  		struct romfs_inode_info		romfs_i;
//  		struct shmem_inode_info		shmem_i;
//  		struct coda_inode_info		coda_i;
//  		struct smb_inode_info		smbfs_i;
//  		struct hfs_inode_info		hfs_i;
//  		struct adfs_inode_info		adfs_i;
//  		struct qnx4_inode_info		qnx4_i;
//  		struct reiserfs_inode_info	reiserfs_i;
//  		struct bfs_inode_info		bfs_i;
//  		struct udf_inode_info		udf_i;
//  		struct ncp_inode_info		ncpfs_i;
//  		struct proc_inode_info		proc_i;
//  		struct socket			socket_i;
//  		struct usbdev_inode_info        usbdev_i;
//  		struct jffs2_inode_info		jffs2_i;
//  		void				*generic_ip;
//  	} u;
} inode;


// linux/fs.h
int register_chrdev(unsigned int, const char *, struct file_operations *);
int unregister_chrdev(unsigned int, const char *);


// linux/sched.h
typedef struct task_struct {
//  	/*
//  	 * offsets of these are hardcoded elsewhere - touch with care
//  	 */
//  	volatile long state;	/* -1 unrunnable, 0 runnable, >0 stopped */
//  	unsigned long flags;	/* per process flags, defined below */
//  	int sigpending;
//  	mm_segment_t addr_limit;	/* thread address space:
//  					 	0-0xBFFFFFFF for user-thead
//  						0-0xFFFFFFFF for kernel-thread
//  					 */
//  	struct exec_domain *exec_domain;
//  	volatile long need_resched;
//  	unsigned long ptrace;

//  	int lock_depth;		/* Lock depth */

//  /*
//   * offset 32 begins here on 32-bit platforms. We keep
//   * all fields in a single cacheline that are needed for
//   * the goodness() loop in schedule().
//   */
//  	long counter;
//  	long nice;
//  	unsigned long policy;
//  	struct mm_struct *mm;
//  	int processor;
//  	/*
//  	 * cpus_runnable is ~0 if the process is not running on any
//  	 * CPU. It's (1 << cpu) if it's running on a CPU. This mask
//  	 * is updated under the runqueue lock.
//  	 *
//  	 * To determine whether a process might run on a CPU, this
//  	 * mask is AND-ed with cpus_allowed.
//  	 */
//  	unsigned long cpus_runnable, cpus_allowed;
//  	/*
//  	 * (only the 'next' pointer fits into the cacheline, but
//  	 * that's just fine.)
//  	 */
//  	struct list_head run_list;
//  	unsigned long sleep_time;

//  	struct task_struct *next_task, *prev_task;
//  	struct mm_struct *active_mm;
//  	struct list_head local_pages;
//  	unsigned int allocation_order, nr_local_pages;

//  /* task state */
//  	struct linux_binfmt *binfmt;
//  	int exit_code, exit_signal;
//  	int pdeath_signal;  /*  The signal sent when the parent dies  */
//  	/* ??? */
//  	unsigned long personality;
//  	int did_exec:1;
  	pid_t pid;
//  	pid_t pgrp;
//  	pid_t tty_old_pgrp;
//  	pid_t session;
//  	pid_t tgid;
//  	/* boolean value for session group leader */
//  	int leader;
//  	/* 
//  	 * pointers to (original) parent process, youngest child, younger sibling,
//  	 * older sibling, respectively.  (p->father can be replaced with 
//  	 * p->p_pptr->pid)
//  	 */
//  	struct task_struct *p_opptr, *p_pptr, *p_cptr, *p_ysptr, *p_osptr;
//  	struct list_head thread_group;

//  	/* PID hash table linkage. */
//  	struct task_struct *pidhash_next;
//  	struct task_struct **pidhash_pprev;

//  	wait_queue_head_t wait_chldexit;	/* for wait4() */
//  	struct completion *vfork_done;		/* for vfork() */
//  	unsigned long rt_priority;
//  	unsigned long it_real_value, it_prof_value, it_virt_value;
//  	unsigned long it_real_incr, it_prof_incr, it_virt_incr;
//  	struct timer_list real_timer;
//  	struct tms times;
//  	unsigned long start_time;
//  	long per_cpu_utime[NR_CPUS], per_cpu_stime[NR_CPUS];
//  /* mm fault and swap info: this can arguably be seen as either mm-specific or thread-specific */
//  	unsigned long min_flt, maj_flt, nswap, cmin_flt, cmaj_flt, cnswap;
//  	int swappable:1;
//  /* process credentials */
//  	uid_t uid,euid,suid,fsuid;
//  	gid_t gid,egid,sgid,fsgid;
//  	int ngroups;
//  	gid_t	groups[NGROUPS];
//  	kernel_cap_t   cap_effective, cap_inheritable, cap_permitted;
//  	int keep_capabilities:1;
//  	struct user_struct *user;
//  /* limits */
//  	struct rlimit rlim[RLIM_NLIMITS];
//  	unsigned short used_math;
//  	char comm[16];
//  /* file system info */
//  	int link_count, total_link_count;
//  	struct tty_struct *tty; /* NULL if no tty */
//  	unsigned int locks; /* How many file locks are being held */
//  /* ipc stuff */
//  	struct sem_undo *semundo;
//  	struct sem_queue *semsleeping;
//  /* CPU-specific state of this task */
//  	struct thread_struct thread;
//  /* filesystem information */
//  	struct fs_struct *fs;
//  /* open file information */
//  	struct files_struct *files;
//  /* signal handlers */
//  	spinlock_t sigmask_lock;	/* Protects signal and blocked */
//  	struct signal_struct *sig;

//  	sigset_t blocked;
//  	struct sigpending pending;

//  	unsigned long sas_ss_sp;
//  	size_t sas_ss_size;
//  	int (*notifier)(void *priv);
//  	void *notifier_data;
//  	sigset_t *notifier_mask;
	
//  /* Thread group tracking */
//     	u32 parent_exec_id;
//     	u32 self_exec_id;
//  /* Protection of (de-)allocation: mm, files, fs, tty */
//  	spinlock_t alloc_lock;

//  /* journalling filesystem info */
//  	void *journal_info;
} task_struct;
struct task_struct *current;    // current process


// asm/errno.h
#define	ENOMEM		12	/* Out of memory */
#define	EFAULT		14	/* Bad address */
#define	EBUSY		16	/* Device or resource busy */
#define	ENODEV		19	/* No such device */
#define	EINVAL		22	/* Invalid argument */
#define	EUCLEAN		117	/* Structure needs cleaning */


// asm/fcntl.h
#define O_ACCMODE	   0003
#define O_RDONLY	     00
#define O_WRONLY	     01
#define O_RDWR		     02
#define O_NONBLOCK	  04000


// asm/uaccess.h

unsigned long copy_to_user(void *to, const void *from, unsigned long n)
  pre(all_of {
    // extract the base and offset/field from the pointer, and check
    // that everything is in bounds; no check on 'to' is necessary
    // since copy_to_user does that check at run-time
    0 <= ptr_field(from) <= ptr_field(from)+n <= objsize[ptr_addr(from)];
  })
;

unsigned long copy_from_user(void *to, const void *from, unsigned long n)
  pre(all_of {
    0 <= ptr_field(to) <= ptr_field(to)+n <= objsize[ptr_addr(to)];
  })
;


#define __copy_from_user copy_from_user
int access_ok(int type, void *addr, int size);
#define VERIFY_READ 0
#define VERIFY_WRITE 1
#define get_user(var, ptr) (var = *(ptr), 0)
#define __get_user(var, ptr) (var = *(ptr), 0)
#define put_user(var, ptr) (*(ptr) = var, 0)


// linux/module.h
#define MODULE_PARM(var, flags)
#define MODULE_PARM_DESC(var, desc)
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT


// asm/ioctl.h
/* ioctl command encoding: 32 bits total, command in lower 16 bits,
 * size of the parameter structure in the lower 14 bits of the
 * upper 16 bits.
 * Encoding the size of the parameter structure in the ioctl request
 * is useful for catching programs compiled with old versions
 * and to avoid overwriting user space outside the user buffer area.
 * The highest 2 bits are reserved for indicating the ``access mode''.
 * NOTE: This limits the max parameter size to 16kB -1 !
 */

/*
 * The following is for compatibility across the various Linux
 * platforms.  The i386 ioctl numbering scheme doesn't really enforce
 * a type field.  De facto, however, the top 8 bits of the lower 16
 * bits are indeed used as a type field, so we might just as well make
 * this explicit here.  Please be sure to use the decoding macros
 * below from now on.
 */
#define _IOC_NRBITS	8
#define _IOC_TYPEBITS	8
#define _IOC_SIZEBITS	14
#define _IOC_DIRBITS	2

#define _IOC_NRMASK	((1 << _IOC_NRBITS)-1)
#define _IOC_TYPEMASK	((1 << _IOC_TYPEBITS)-1)
#define _IOC_SIZEMASK	((1 << _IOC_SIZEBITS)-1)
#define _IOC_DIRMASK	((1 << _IOC_DIRBITS)-1)

#define _IOC_NRSHIFT	0
#define _IOC_TYPESHIFT	(_IOC_NRSHIFT+_IOC_NRBITS)
#define _IOC_SIZESHIFT	(_IOC_TYPESHIFT+_IOC_TYPEBITS)
#define _IOC_DIRSHIFT	(_IOC_SIZESHIFT+_IOC_SIZEBITS)

/*
 * Direction bits.
 */
#define _IOC_NONE	0U
#define _IOC_WRITE	1U
#define _IOC_READ	2U

#define _IOC(dir,type,nr,size) \
	(((dir)  << _IOC_DIRSHIFT) | \
	 ((type) << _IOC_TYPESHIFT) | \
	 ((nr)   << _IOC_NRSHIFT) | \
	 ((size) << _IOC_SIZESHIFT))

/* used to create numbers */
#define _IO(type,nr)		_IOC(_IOC_NONE,(type),(nr),0)
#define _IOR(type,nr,size)	_IOC(_IOC_READ,(type),(nr),sizeof(size))
#define _IOW(type,nr,size)	_IOC(_IOC_WRITE,(type),(nr),sizeof(size))
#define _IOWR(type,nr,size)	_IOC(_IOC_READ|_IOC_WRITE,(type),(nr),sizeof(size))

/* used to decode ioctl numbers.. */
#define _IOC_DIR(nr)		(((nr) >> _IOC_DIRSHIFT) & _IOC_DIRMASK)
#define _IOC_TYPE(nr)		(((nr) >> _IOC_TYPESHIFT) & _IOC_TYPEMASK)
#define _IOC_NR(nr)		(((nr) >> _IOC_NRSHIFT) & _IOC_NRMASK)
#define _IOC_SIZE(nr)		(((nr) >> _IOC_SIZESHIFT) & _IOC_SIZEMASK)

/* ...and for the drivers/sound files... */

#define IOC_IN		(_IOC_WRITE << _IOC_DIRSHIFT)
#define IOC_OUT		(_IOC_READ << _IOC_DIRSHIFT)
#define IOC_INOUT	((_IOC_WRITE|_IOC_READ) << _IOC_DIRSHIFT)
#define IOCSIZE_MASK	(_IOC_SIZEMASK << _IOC_SIZESHIFT)
#define IOCSIZE_SHIFT	(_IOC_SIZESHIFT)


// unknown
#define MINOR(dev) (dev & 0xFF)
#define EXPORT_NO_SYMBOLS()


// ------------------ from mykmalloc.h --------------------
#define MYKMALLOCPRINTK(size) malloc(size)
#define mykfree(p) kfree(p)
#define mykfreesz(p,sz) kfree(p)


// -------------------- from kutil.h -------------------
// return type of functions which return 0 or positive for
// success, and negative errno values for failure; other types
// that often have these semantics include ssize_t and loff_t
typedef int Err;

// integer as boolean
typedef int bool;
//enum { false=0, true=1 };
#define false 0
#define true 1


// probably the most ubiquitous macro...
#ifndef min
  #define min(a,b) ((a)<(b) ? (a) : (b))
#endif
  

// printk that includes the process id of the associated user process
#define printkpid(kernLevel, fmt, args...)                    \
  printk(kernLevel "[%d] " fmt, current->pid , ## args);      \

  
// standard debugging flags; standard flags will be capital
// letters, and module-specific ones will be lowercase
// (mnemonic for all letters: SYLBOG)
enum {
  // this end: generally useful
  DF_LEAKS = 'L',               // report memory leaks at unload time
  DF_OPEN_RELEASE = 'O',        // calls to open and release
  DF_SYSCALL = 'Y',             // system calls other than open and release
  DF_SIGNAL = 'G',              // signal activity
  DF_SLEEP = 'S',               // sleep/wakeup
  DF_BADIOCTL = 'B',            // print when an unknown ioctl is issued
  // this end: for debugging
};


#define ERRCALL(call)  \
  err = call;          \
  if (err < 0) {       \
    return err;        \
  }
    
  
  // NDEBUG branch
  #define debugFlagIsSet(flag) false
  #define dprintk(flag, fmt, args...) ((void)0);
  #define dprintkpid(flag, fmt, args...) ((void)0);

  #define CHECK(condition, codeIfFalse)      \
    if (!(condition)) {                      \
      return codeIfFalse;                    \
    }


// ------------------ from scull.h ------------------
// macros to extract the service number ("type") and device number
// ("num") from the minor device number
#define TYPE(dev)   (MINOR(dev) >> 4)     // high nibble
#define NUM(dev)    (MINOR(dev) & 0xF)    // low nibble


// the primary scull data structure: this is a node in a linked list.
// each node stores a portion of the total data in the clipboard;
// for each scull device, filp->private_data points at one of these
typedef struct Scull_Dev
{
  // ---- per-device data; only meaningful in the *first* node ----
  int quantum;              /* the current quantum size in bytes */
  int qset;                 /* the current quantum set array size, in # of pointers */

  // this appears to be the total size of the data stored in all list
  // nodes, in bytes; I think it is only maintained for the first list
  // node, and remains uninitialized for the rest
  unsigned long totsize;

  unsigned int access_key;  /* used by sculluid and scullpriv (to do what?) */

  // lock the device while using it; this must be nonzero if we ever
  // might call schedule() and then resume accessing the data on
  // wakeup
  unsigned int usage;

  // place for ioctl to store an int
  int ioctl_scratch_value;

  // keep track of whether we're in blocking or nonblocking mode
  bool nonblocking;

  // ---- per-quantum-set data; meaningful in *all* nodes ----
  // this is a pointer to a "quantum set", which is itself an
  // array of (up to) 1000 pointers to "quanta" (4000-byte blocks)
  char **data;              // (void * owner * owner)

  // pointer to next item in the list
  struct Scull_Dev *next;   // (owner)

  // instance whose 'next' points to me, if any
  struct Scull_Dev *prev;   // ghost

  // true if this is a first node, or one of the nodes that follows on
  // after the first
  bool first;

} Scull_Dev;


// things to go in file_operations for 'scull'
Err scull_open(struct inode *inode, struct file *filp);
Err scull_release(struct inode *inode, struct file *filp);
loff_t scull_lseek(struct file *filp, loff_t off, int whence);
Err scull_ioctl(struct inode *inode, struct file *filp,
                unsigned int cmd, unsigned long arg);
ssize_t scull_read(struct file *filp, char *buf, size_t count,
                   loff_t *f_pos);
ssize_t scull_write(struct file *filp, const char *buf, size_t count,
                    loff_t *f_pos);


// state of the entire module
enum ModuleState { MS_UNLOADED, MS_LOADED } module_state = MS_UNLOADED;


// internal helpers
Err scull_trim(Scull_Dev *dev)
  pre(all_of {
    module_state==MS_LOADED;
    0 < dev < objct;        tag[dev]==Scull_Dev_tag;
    dev->first == true;
  })
  post(all_of {
    result <= 0;
    result==0 ==> dev->next == NULL;

    // monotonically approaching NULL 'next' pointers
    forall(Scull_Dev *p).
      tag[p]==Scull_Dev_tag &&
      pre(p->next)==NULL
    ==>
      p->next==NULL;

    // no modification of any of the struct 'file's
    forall(file *p). tag[p]==file_tag <==> pre(tag)[p]==file_tag;
    //forall(file *p, int f). tag[p]==file_tag ==> mem[p,f] == pre(mem)[p,f];  // just don't mention in 'changed'
  })
  changes(mem, meta, tag, Scull_Dev_prev, Scull_Dev_data, Scull_Dev_next,
          Scull_Dev_totsize,               
          
          // these are written, but to the same value that they already
          // have; but to show they're not actually changed I'd need to
          // carry in the loop invariants the fact that the pre-state
          // values are equal to scull_quantum/qset.  I think fact flow
          // would get that automatically if I had it hooked up.
          Scull_Dev_quantum, Scull_Dev_qset)
;

Scull_Dev *scull_follow(Scull_Dev *dev, int n)
  pre(module_state==MS_LOADED &&
      0 < dev < objct &&
      tag[dev]==Scull_Dev_tag &&
      n >= 0)
  post(all_of {
    0 < result < objct;
    tag[result]==Scull_Dev_tag;
    
    // nothing in 'tag' below the allocation boundary changed
    forall(void *p). 0<p<pre(objct) ==> tag[p] == pre(tag)[p];
    
    // objct only increases when it changes
    objct >= pre(objct);
  })
  changes(mem, meta, tag, objsize, objct,
          Scull_Dev_tag /*all Scull fields*/)
;

void check_nonblock_change(struct file *filp, char const *caller);


// init/deinit for scullpipe
Err init_pipe_module();
void cleanup_pipe_module();


// first, the "magic number" that is supposed to be only used by
// this driver; since that's impossible (there are too many drivers),
// the code serves to spread out the space somewhat, in hopes of
// reducing the likelihood of accidentally issuing the right command
// to the wrong device
#define SCULL_IOC_MAGIC 'k'

// tells the driver to do a printk of its memory allocation statistics
#define SCULL_MALLOC_STATS _IO(SCULL_IOC_MAGIC, 1)

// passes an integer value via pointer into the driver, where it just
// remembers the value (here, "read" and "write" are from the user's
// perspective); NOTE: the 3rd arg to _IO{W,R} has an implicit 'sizeof'
// wrapped around it
#define SCULL_STORE_VALUE _IOW(SCULL_IOC_MAGIC, 2, /*sizeof*/ int)

// retrieves the integer value stored by SCULL_STORE_VALUE; the value
// is initially 0
#define SCULL_GET_VALUE _IOR(SCULL_IOC_MAGIC, 3, /*sizeof*/ int)

// like STORE_VALUE, but the driver doesn't validate the address!
// (the code has direction of 0, so this will bypass the check)
#define SCULL_STORE_VALUE_UNCHECKED _IO(SCULL_IOC_MAGIC, 4)


// define the range of the command codes
#define SCULL_MIN_CODE 1
#define SCULL_MAX_CODE 4


// --------------- from main.c ------------------
// for printk calls
#define SCULL "scull: "
#define SCULL_INFO KERN_INFO SCULL
#define SCULL_WARNING KERN_WARNING SCULL
#define SCULL_ERR KERN_ERR SCULL


// the driver implements several services:
//
// scull: like a kernel clipboard: apps can store a block of data into
// it, and then read it out (nondestructively) later
//
// scullpipe: same as a fifo
//
// scullsingle: same as scull, but only one process at a time
//
// scullpriv: same as scull, but each virtual console gets its own
// clipboard
//
// sculluid: ?
// scullwuid: same as sculluid, but blocks when 2nd process opens (?)


// --------------------- global data -------------------------
// file operations for the basic 'scull' service
struct file_operations scull_fops = {
  scull_lseek,
  scull_read,
  scull_write,
  NULL,          /* scull_readdir */
  NULL,          /* scull_select */
  scull_ioctl,
  NULL,          /* scull_mmap */
  scull_open,
  NULL,          // flush
  scull_release,
                 /* nothing more, fill with NULLs */
};


// array of pointers to fn blocks, one for each type of service
struct file_operations *scull_fop_array[] = {
  &scull_fops,              // type 0: scull (clipboard)
#if 0     // not yet
  &scull_pipe_fops,         //      1: fifo pipe
  &scull_priv_fops,
  &scull_sngl_fops,
  &scull_user_fops,
  &scull_wusr_fops,
#endif // 0
};
//#define SCULL_MAX_TYPE 5
#define SCULL_MAX_TYPE 0


// user-configurable parameters and their default values
int scull_major = 0;        // default: dynamic assignment
MODULE_PARM(scull_major, "i");
MODULE_PARM_DESC(scull_major, "major device number; 0 for dynamic assignment");

int scull_nr_devs = 4;
MODULE_PARM(scull_nr_devs, "i");
MODULE_PARM_DESC(scull_nr_devs, "number of devices per service; in [1,4]");

int scull_quantum = 4000;
MODULE_PARM(scull_quantum, "i");
MODULE_PARM_DESC(scull_quantum, "size in bytes of a quantum; >0");

int scull_qset = 1000;
MODULE_PARM(scull_qset, "i");
MODULE_PARM_DESC(scull_qset, "number of quanta in a quantum set; >0");

char *debug = "OGL";
MODULE_PARM(debug, "s");
MODULE_PARM_DESC(debug,
  "string of debug flags; default is \"OGL\"");

// debugging flags (beyond those defined in kutil.h)
enum {
  DF_ACCESS = 'a',       // various tests with access_ok
  DF_SCRATCH = 's',      // report scratch value exchanges
  DF_NONBLOCK = 'n',     // print detected changes in O_NONBLOCK
};


// array of first list nodes, allocated in init_module
Scull_Dev *scull_devices = NULL;   // (owner)


// id of this module
#define SCULL_MODULE_ID 22

// user-defined type tags
#define L1TAG 101        // tag on array to which 'data' points
#define L2TAG 102        // tag on arrays pointed to by 'data[i]'
#define ZEROED_SCULL_DEV_TAG 103    // Scull_Dev with all zero fields


// maps that give parent information for the arrays
int L1parent;           // maps L1 addresses to Scull_Dev addresses
int L2parent;           // maps L2 addresses to L1 addresses
int L2index;            // maps L2 addresses to the index that points to them in the parent L1 array


// ------------------- invariants ---------------------
// 'dev' is one of the toplevel devices
#define isToplevelDevice(dev) \
  (scull_devices <= (dev) < scull_devices+scull_nr_devs)


// invariants about 'dev', a Scull_Dev instance
int Scull_Dev_invariants();
define Scull_Dev_invariants(Scull_Dev *dev, int mem, int tag, int objct, 
                            int scull_devices, int scull_nr_devs) {
  tag[dev]==Scull_Dev_tag;

  // in the original code, the invariant only held for the first
  // element of the list, but I've modified it so it holds for all
  // elements
  dev->quantum == scull_quantum;
  dev->qset == scull_qset;

  dev->data!=NULL ==> all_of {
    0 < dev->data < objct;
    tag[dev->data] == L1TAG;

    // parent pointer is accurate
    L1parent[dev->data] == dev;
  };

  dev->next!=NULL ==> all_of {
    0 < dev->next < objct;
    tag[dev->next] == Scull_Dev_tag;

    dev->next != dev;
    dev->next->prev == dev;
    dev->next->first == false;    // 'first' is my new marker for 'toplevel'
  };

  dev->prev!=NULL ==> all_of {
    0 < dev->prev < objct;
    tag[dev->prev] == Scull_Dev_tag;

    dev->prev != dev;
    dev->prev->next == dev;
  };
  
  dev->first==true ==> dev->prev==NULL;
}

                         
// bit of a hack, until I bite the bullet and allow casts
// into the VC expr language
int firstIsTrue(Scull_Dev *dev);
define firstIsTrue(Scull_Dev *dev) {
  dev->first == true;
}


int global_inv() { return vc_expression(all_of {
  0 < objct;
  tag[NULL] == 0;

  0 < scull_nr_devs;
  0 < scull_quantum;
  0 < scull_qset;

  module_state==MS_UNLOADED || module_state==MS_LOADED;

  // if the module is not loaded, 'scull_devices' is NULL
  module_state==MS_UNLOADED ==> scull_devices==NULL;

  // if the module is loaded, then 'scull_devices' is a valid pointer
  // to the beginning of 'scull_nr_devs' contiguous Scull_Devs
  module_state==MS_LOADED ==> all_of {
    0 < scull_devices < scull_devices+scull_nr_devs <= objct;

    // here, I'd at one point restricted the quantification using
    // "isToplevelDevice", but because of Simplify's restrictive
    // pattern matching on arithmetical constructs, it's better
    // to use an explicit 'scull_devices+i' approach
    forall(int i). 0<=i<scull_nr_devs ==> all_of {
      tag[scull_devices+i]==Scull_Dev_tag;                                          // disequality
      (scull_devices+i)->first == true;
    };
  };

  // -- invariants about instances of 'struct file' --

  // either closed or open
  forall(file *filp) PATS(filp->state). tag[filp]==file_tag ==>
    filp->state==FS_CLOSED || filp->state==FS_OPEN;

  // if open and associated with this module, then the private_data
  // field is in fact of type Scull_Dev whose 'prev' is NULL
  forall(file *filp) PATS(filp->private_data). tag[filp]==file_tag ==>
    filp->state==FS_OPEN && filp->id==SCULL_MODULE_ID ==> (
      0 < filp->private_data < objct &&
      tag[filp->private_data] == Scull_Dev_tag &&                                   // disequality
      //((Scull_Dev*)filp->private_data)->first == true
      firstIsTrue(filp->private_data)
    );

  // -- invariants about instances of 'struct Scull_Dev' --
  //forall(Scull_Dev *dev). tag[dev]==Scull_Dev_tag ==>
  //  Scull_Dev_invariants(dev, mem, tag, objct, scull_devices, scull_nr_devs);

  forall(Scull_Dev *dev) PATS(dev->quantum). tag[dev]==Scull_Dev_tag ==>
    dev->quantum == scull_quantum;
  forall(Scull_Dev *dev) PATS(dev->qset). tag[dev]==Scull_Dev_tag ==>
    dev->qset == scull_qset;

  forall(Scull_Dev *dev) PATS(dev->data). tag[dev]==Scull_Dev_tag ==>
    dev->data!=NULL ==> all_of {
      0 < dev->data < objct;
      tag[dev->data] == L1TAG;

      // parent pointer is accurate
      L1parent[dev->data] == dev;                                                   // inj
    };

  forall(Scull_Dev *dev) PATS(dev->next). tag[dev]==Scull_Dev_tag ==>
    dev->next!=NULL ==> all_of {
      0 < dev->next < objct;
      tag[dev->next] == Scull_Dev_tag;

      dev->next != dev;                                                             // !=
      dev->next->prev == dev;                                                       // inj
      dev->next->first == false;    // 'first' is my new marker for 'toplevel'
    };

  forall(Scull_Dev *dev) PATS(dev->prev). tag[dev]==Scull_Dev_tag ==>
    dev->prev!=NULL ==> all_of {
      0 < dev->prev < objct;
      tag[dev->prev] == Scull_Dev_tag;

      dev->prev != dev;                                                             // !=
      dev->prev->next == dev;                                                       // inj
    };

  forall(Scull_Dev *dev) PATS(dev->prev/*?*/). tag[dev]==Scull_Dev_tag ==>
    dev->first==true ==> dev->prev==NULL;


  // invariants about instances of level 1 arrays
  forall(char **data). tag[data]==L1TAG ==> all_of {
    objsize[data] == scull_qset;       // size is fixed
    forall(int i). 0<=i<scull_qset && mem[data,i]!=0 ==> all_of {
      0 < mem[data,i] < objct;
      tag[mem[data,i]]==L2TAG;

      L2parent[mem[data,i]] == data;                                                // inj
      L2index[mem[data,i]] == i;                                                    // inj
    };
  };

  // invariants about instances of level 2 arrays
  forall(char *p). tag[p]==L2TAG ==>
    objsize[p]==scull_quantum;         // size is fixed

});}


// --------------------- functions -------------------------
#define DO_SCULL_OPEN 1                // ok
#define DO_SCULL_RELEASE 1             // ok
#define DO_SCULL_TRIM 1                // ok
#define DO_SCULL_READ 1                // ok
#define DO_SCULL_WRITE 1               // ok
#define DO_SCULL_FOLLOW 1              // ok
#define DO_SCULL_LSEEK 0               // never
#define DO_USERPTR_EXPERIMENT 0        // never
#define DO_SCULL_IOCTL 0               // never
#define DO_CHECK_NONBLOCK_CHANGE 0     // never
#define DO_INIT_MODULE 1               // ok
#define DO_CLEANUP_MODULE 1            // ok


#if DO_SCULL_OPEN
// the initial open(2) system call is handled by this
Err scull_open(struct inode *inode, struct file *filp)
  pre(all_of {
    module_state==MS_LOADED;
    0 < inode < objct;      tag[inode] == inode_tag;
    0 < filp < objct;       tag[filp] == file_tag;
    filp->state==FS_CLOSED;
    filp->id==SCULL_MODULE_ID;
  })
  post(all_of {
    result <= 0;
    result==0 ==> filp->state==FS_OPEN;
    result<0  ==> filp->state==FS_CLOSED;
  })
  changes(mem/*?*/, meta/*?*/, tag, file_state, file_f_op, file_private_data,
          Scull_Dev_tag /*all Scull fields*/)
{
  assert(filp->state == FS_CLOSED);

  // which service to implement?
  int type = TYPE(inode->i_rdev);

  // and which instance (there are 4) of the service?
  int num = NUM(inode->i_rdev);

  Scull_Dev *dev;     // device information (?)

  // hack: special minor number as module count escape hatch
  if (MINOR(inode->i_rdev) == 255) {
    printk(SCULL_INFO "manually decrementing module count...\n");
    MOD_DEC_USE_COUNT;
    return -EUCLEAN;     // :)
  }

  if (type != 0) {
    // services other than the basic clipboard
    if (type < 0 || type > SCULL_MAX_TYPE) {
      // bad minor device number
      printk(SCULL_ERR "Minor device number is invalid: %d\n",
             MINOR(inode->i_rdev));
      return -ENODEV;
    }

    // NOTE: this path is infeasible for now since I've set
    // SCULL_MAX_TYPE to 0
    assert(1==0);

    // replace default dispatch table with service-specific table
    filp->f_op = scull_fop_array[type];

    // and re-dispatch open call
    return filp->f_op->open(inode, filp);
  }

  // -- service type 0: clipboard --

  // check the device number; the only way this fails is if someone
  // sets scull_nr_devs to less than 4
  if (num >= scull_nr_devs) {
    printk(SCULL_ERR "Invalid device number: %d\n", num);
    return -ENODEV;
  }

  // point 'dev' at the particular instance being accessed
  assert(0 <= num < scull_nr_devs);
  dev = &scull_devices[num];

  // grab initial value of blocking flag
  if (filp->f_flags & O_NONBLOCK) {
    dev->nonblocking = 1;
    dprintkpid(DF_OPEN_RELEASE, SCULL "device opened in nonblocking mode\n");
  }
  else {
    dev->nonblocking = 0;
    dprintkpid(DF_OPEN_RELEASE, SCULL "device opened in blocking mode\n");
  }

  // if the device is opened for write-only, we truncate the data
  // to 0 length
  if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
    // ignore errors; an error might arise due to an anticipated
    // reentrancy issue (which I don't fully understand yet)
    scull_trim(dev);
  }

  // associate this open file structure with the device instance
  // in a way that makes it convenient to retrieve later
  filp->private_data = dev;

  // mark the file as being open
  filp->state = FS_OPEN;

  // the module is now in use, so it can't be unloaded
  MOD_INC_USE_COUNT;

  // success
  return 0;
}
#endif // DO_SCULL_OPEN


#if DO_SCULL_RELEASE
// this is called in response to the close(2) system call
// for the 'scull' service; other services have their own
// release() functions, installed by scull_open above
Err scull_release(struct inode *inode, struct file *filp)
  pre(all_of {
    module_state==MS_LOADED;
    0 < inode < objct;      tag[inode] == inode_tag;
    0 < filp < objct;       tag[filp] == file_tag;
    filp->state==FS_OPEN;
    filp->id==SCULL_MODULE_ID;
  })
  post(all_of {
    result <= 0;
    filp->state==FS_CLOSED;     // regardless of result
  })
  changes(file_state)
{
  check_nonblock_change(filp, "release");

  // we don't deallocate the clipboard; it remains available
  // to other apps

  // there is physical device to shut down

  dprintkpid(DF_OPEN_RELEASE, SCULL "device closed\n");

  filp->state = FS_CLOSED;

  // so only thing to do is decrement usage count, such that
  // someone could unload the module
  MOD_DEC_USE_COUNT;

  // success
  return 0;
}
#endif // DO_SCULL_RELEASE


#if DO_SCULL_TRIM
// this is called to truncate the stored data to 0 length
Err scull_trim(Scull_Dev *dev)
  pre(all_of {
    module_state==MS_LOADED;
    0 < dev < objct;        tag[dev]==Scull_Dev_tag;
    dev->first == true;
  })
  post(all_of {
    result <= 0;
    result==0 ==> dev->next == NULL;

    // monotonically approaching NULL 'next' pointers
    forall(Scull_Dev *p).
      tag[p]==Scull_Dev_tag &&
      pre(p->next)==NULL
    ==>
      p->next==NULL;

    // no modification of any of the struct 'file's
    forall(file *p). tag[p]==file_tag <==> pre(tag)[p]==file_tag;
    //forall(file *p, int f). tag[p]==file_tag ==> mem[p,f] == pre(mem)[p,f];  // just don't mention in 'changed'
  })
  changes(mem, meta, tag, Scull_Dev_prev, Scull_Dev_data, Scull_Dev_next,
          Scull_Dev_totsize,               
          
          // these are written, but to the same value that they already
          // have; but to show they're not actually changed I'd need to
          // carry in the loop invariants the fact that the pre-state
          // values are equal to scull_quantum/qset.  I think fact flow
          // would get that automatically if I had it hooked up.
          Scull_Dev_quantum, Scull_Dev_qset)
{
  Scull_Dev *next, *dptr;     // for walking the linked list
  int qset = dev->qset;       // 'dev' is always non-null
  int i;

  if (dev->usage) {
    printkpid(SCULL_WARNING, "Attempt to trim scull data while another process "
              "is accessing it.\n");
    return -EBUSY; /* scull_open ignores this error and goes on */
  }

  for (dptr = dev; dptr; dptr = next) {             /* all the list items */
    invariant {
      // precondition
      global_inv;
      0 < dev < objct;     tag[dev]==Scull_Dev_tag;

      // monotonically approaching NULL 'next' pointers
      forall(Scull_Dev *p).
        tag[p]==Scull_Dev_tag &&
        pre(p->next)==NULL
      ==>
        p->next==NULL;

      // no modification of any of the struct 'file's
      forall(file *p). tag[p]==file_tag <==> pre(tag)[p]==file_tag;
      //forall(p,f). tag[p]==file_tag ==> mem[p,f] == pre(mem)[p,f];

      // equality from above
      qset == scull_qset;

      // this loop
      0 < dptr < objct;    tag[dptr]==Scull_Dev_tag;
      dptr->prev == NULL;     // no Scull_Dev is pointing at dptr
      dptr == dev || (dptr->first == false &&
                      dev->next == NULL);
    }

    if (dptr->data) {
      for (i = 0; i < qset; i++) {                  // for each quantum..
        invariant {
          // precondition
          global_inv;
          0 < dev < objct;     tag[dev]==Scull_Dev_tag;

          // monotonically approaching NULL 'next' pointers
          forall(Scull_Dev *p).
            tag[p]==Scull_Dev_tag &&
            pre(p->next)==NULL
          ==>
            p->next==NULL;

          // no modification of any of the struct 'file's
          forall(file *p). tag[p]==file_tag <==> pre(tag)[p]==file_tag;
          //forall(p,f). tag[p]==file_tag ==> mem[p,f] == pre(mem)[p,f];

          // equality from above
          qset == scull_qset;

          // outer loop
          0 < dptr < objct;    tag[dptr]==Scull_Dev_tag;
          dptr->prev == NULL;     // no Scull_Dev is pointing at dptr
          dptr == dev || (dptr->first == false &&
                          dev->next == NULL);

          // guard around loop
          dptr->data != NULL;

          // this loop
          0 <= i < qset;
        }

        if (dptr->data[i]) {                        // ..if it's there..
          mykfreesz(dptr->data[i], dev->quantum);   // ..free it
          dptr->data[i] = NULL;                     // restore global inv
        }
      }

      mykfreesz(dptr->data, sizeof(char*) * qset);  // free the quantum set
      dptr->data=NULL;                              // this node now has no quantum set
    }

    assert(
      dptr == dev || (dptr->first == false &&
                      dev->next == NULL)
    );

    next = dptr->next;                              // prepare to advance to next item
    if (next) {
      assert(next->first == false);
      next->prev = NULL;                            // ghost
    }

    if (dptr != dev) {                              /* all of them but the first */
      //assume(metaVar(dptr) == live);                // hack.. don't want to do live/dead now
      mykfreesz(dptr, sizeof(Scull_Dev));           // free a list node
      assert(dev->next == NULL);
    }
    else {
      // moved from below so that the global invariant remains true..
      // otherwise I'd need to say something like "global_inv except ..."
      // and I don't have a way to do that now
      dev->next = NULL;
    }
  }

  dev->totsize = 0;                    // total size is now 0

  // why do we reset these?  what might have changed them?
  dev->quantum = scull_quantum;
  dev->qset = scull_qset;

  // update: do this again.  it's easier than adding the null'd 'next'
  // to the loop invariant above
  // update: but just setting something to NULL makes the verifier
  // unhappy because it wonders about making something dangle
  assert(dev->next == NULL);      // there is no next item now

  // expansion of part of the global invariant
  Scull_Dev *p;
  assume(tag[p]==Scull_Dev_tag);
  assert(Scull_Dev_invariants(p, mem, tag, objct, scull_devices, scull_nr_devs));

  // success
  return 0;
}
#endif // DO_SCULL_TRIM


#if DO_SCULL_READ
// this is called in response to the read(2) system call, when
// applied to an instance of the basic 'scull' service
ssize_t scull_read(
  struct file *filp,         // open-file structure in calling process' area
  char *buf,                 // user-space pointer to destination buffer
  size_t count,              // max # of bytes to copy to 'buf'
  loff_t *ignore_f_pos)      // redundant pointer to filp->f_pos

  pre(all_of {
    module_state==MS_LOADED;
    0 < filp < objct;       tag[filp] == file_tag;
    filp->state==FS_OPEN;
    filp->id==SCULL_MODULE_ID;
    0 <= count;      // is size_t usually 'unsigned'? 
  })
  changes(mem/*no*/, meta/*no*/, tag, objsize, objct, file_f_pos,
          Scull_Dev_tag /*all Scull fields*/)
{
  // grab first element of list associated with this instance
  Scull_Dev *dev = filp->private_data; /* the first listitem */

  int quantum = dev->quantum;
  int qset = dev->qset;

  // this is the maximum # of bytes in this node of the list
  int itemsize = quantum * qset; /* how many bytes in the listitem */
  assert(itemsize > 0);

  // caller's current position in the file
  unsigned long f_pos = (unsigned long)(filp->f_pos);
  assume(f_pos >= 0);    // a type-based fact...

  int item, s_pos, q_pos, rest, resULT;    // name 'result' is reserved by verifier...

  dprintkpid(DF_SYSCALL, SCULL "read, count=%d\n", count);
  
  check_nonblock_change(filp, "read");

  if (f_pos > dev->totsize) {
    // the user tried to read beyond the end of the data
    // (shouldn't the test be ">="?)
    return 0;
  }

  if (f_pos + count > dev->totsize) {
    // if caller requested more data than is available, reduce the
    // count here to pretend the caller only wanted what *is* avail.
    count = dev->totsize - f_pos;
  }

  /* find listitem, qset index and offset in the quantum */
  item = f_pos / itemsize;      // which list node
  rest = f_pos % itemsize;      // offset into the list node
  s_pos = rest / quantum;       // which quatum in the quantum set
  q_pos = rest % quantum;       // offset into the quantum

  assert(0 <= rest < itemsize);
  assert(0 <= s_pos < qset);
  assert(0 <= q_pos < quantum);

  /* follow the list up to the right position (defined elsewhere) */
  dev = scull_follow(dev, item);
  if (!dev) {                               
    // not possible because we only have to allocate inside 'scull_follow'
    // if we're moving to an offset beyond 'size'
    printkpid(KERN_WARNING, "scull_follow failed during a read, which is "
              "not supposed to be possible\n");
    return -ENOMEM;
  }

  // assuming the above relational test were changed to ">=", I
  // don't see how either of these could fail.. reentrancy issue?
  if (!dev->data)
    return 0; /* don't fill holes */
  if (!dev->data[s_pos])
    return 0;

  if (count > quantum - q_pos)
    count = quantum - q_pos; /* read only up to the end of this quantum */

  dev->usage++; /* the following call may sleep */
  resULT = copy_to_user(buf,                      // user-space destination
                        dev->data[s_pos]+q_pos,   // kernel-space source
                        count);                   // # of bytes to transfer
  if (resULT > 0) {
    // user area isn't accessible
    printkpid(SCULL_ERR, "copy_to_user reports %d of %d bytes uncopyable\n",
              resULT, count);
    dev->usage--;        // decrement usage because we're safely out of there
    return -EFAULT;      // but don't advance the file position counter
  }
  dev->usage--;

  filp->f_pos += count;
  return count;
}
#endif // DO_SCULL_READ


#if DO_SCULL_WRITE
// this is called in response to write(2) syscall
ssize_t scull_write(
  struct file *filp,          // open-file in process area
  const char *buf,            // user-space pointer to source data
  size_t count,               // amount of data in 'buf'
  loff_t *ignore_f_pos)       // redundant pointer to filp->f_pos
  
  pre(all_of {
    module_state==MS_LOADED;
    0 < filp < objct;       tag[filp] == file_tag;
    filp->state==FS_OPEN;
    filp->id==SCULL_MODULE_ID;
    0 <= count;      // is size_t usually 'unsigned'? 
  })
  changes(mem, meta, tag, objsize, objct, L1parent, L2parent, L2index,
          file_f_pos, Scull_Dev_tag /*all Scull fields*/)
{
  Scull_Dev *dev = filp->private_data;     // list associated with this instance
  Scull_Dev *dptr;                         // for looking at subsequent list nodes

  int quantum = dev->quantum;
  int qset = dev->qset;
  int itemsize = quantum * qset;           // bytes per list node

  unsigned long f_pos = (unsigned long)(filp->f_pos);
  assume(f_pos >= 0);    // a type-based fact...

  int item, s_pos, q_pos, rest, resULT;

  dprintkpid(DF_SYSCALL, SCULL "write, count=%d\n", count);

  check_nonblock_change(filp, "write");

  /* find listitem, qset index and offset in the quantum */
  item = f_pos / itemsize;
  rest = f_pos % itemsize;
  s_pos = rest / quantum;
  q_pos = rest % quantum;

  assert(0 <= rest < itemsize);
  assert(0 <= s_pos < qset);
  assert(0 <= q_pos < quantum);

  /* follow the list up to the right position */
  dptr = scull_follow(dev, item);
  if (!dptr) {
    // failed to allocate inside scull_follow
    return -ENOMEM;
  }

  assert {
    //global_inv;
    0 < dev < objct;
    tag[dev]==Scull_Dev_tag;
  }

  if (!dptr->data) {
    // we want to write data to this list node, but it doesn't have
    // a quantum set array yet
    dptr->data = (char**)MYKMALLOCPRINTK(qset * sizeof(char *));
    if (!dptr->data) {
      return -ENOMEM;
    }
    char **newBlock = dptr->data;
    tag = vc_expression(tag{newBlock := L1TAG});
    L1parent = vc_expression(L1parent{newBlock := dptr});

    // initially clear the quantum set
    // (why is this necessary?  we will only read data that was written
    // by a call to memcpy_fromfs...)
    memset(dptr->data, 0, qset * sizeof(char *));
  }

  assert {
    //global_inv;
    0 < dev < objct;
    tag[dev]==Scull_Dev_tag;
    dptr->data!=NULL;
  }

  if (!dptr->data[s_pos]) {
    // want to write data into the 's_pos' quantum, but it doesn't exist yet
    dptr->data[s_pos] = (char*)MYKMALLOCPRINTK(quantum * sizeof(char));
    if (!dptr->data[s_pos]) {
      return -ENOMEM;
    }

    char *newBlock2 = dptr->data[s_pos];
    tag = vc_expression(tag{newBlock2 := L2TAG});
    L2parent = vc_expression(L2parent{newBlock2 := dptr->data});
    L2index = vc_expression(L2index{newBlock2 := s_pos});
  }

  assert {
    global_inv;
    0 < dev < objct;
    tag[dev]==Scull_Dev_tag;
    dptr->data!=NULL;
    mem[dptr->data,s_pos]!=NULL;
  }

  if (count > quantum - q_pos)
    count = quantum - q_pos; /* write only up to the end of this quantum */

  dev->usage++; /* the following call may sleep */
  resULT = copy_from_user(dptr->data[s_pos]+q_pos,     // kernel-space destination
                          buf,                         // user-space source
                          count);                      // # of bytes to copy
  if (resULT > 0) {
    // user area isn't accessible
    printkpid(SCULL_ERR, "copy_from_user reports %d of %d bytes uncopyable\n",
              resULT, count);
    dev->usage--;        // decrement usage because we're safely out of there
    return -EFAULT;      // but don't advance the file position counter or size
  }
  dev->usage--;

  /* update the size */
  if (dev->totsize < f_pos + count)
    dev->totsize = f_pos + count;       // the write made the data area bigger

  filp->f_pos += count;              // advance user's position pointer

  return count;                      // success
}
#endif // DO_SCULL_WRITE


#if DO_SCULL_FOLLOW
/*
 * Follow the list
 */
Scull_Dev *scull_follow(Scull_Dev *dev, int n)
  // NOTE: this signature is repeated above at the prototype; they must
  // be kept the same for soundness!
  pre(module_state==MS_LOADED &&
      0 < dev < objct &&
      tag[dev]==Scull_Dev_tag &&
      n >= 0)
  post(all_of {
    0 < result < objct;
    tag[result]==Scull_Dev_tag;

    // nothing in 'tag' below the allocation boundary changed
    forall(void *p). 0<p<pre(objct) ==> tag[p] == pre(tag)[p];

    // objct only increases when it changes
    objct >= pre(objct);
  })
  changes(mem, meta, tag, objsize, objct,
          Scull_Dev_tag /*all Scull fields*/)
{
  while (n--) {
    invariant {
      // precondition
      global_inv;
      0 < dev < objct;
      tag[dev]==Scull_Dev_tag;
      forall(void *p). 0<p<pre(objct) ==> tag[p] == pre(tag)[p];
      objct >= pre(objct);
                     
      // this loop
      n >= 0;     // not necessary unless I'm worrying about termination
    }

    if (!dev->next) {
      // the original code did not check kmalloc's return value, which
      // could definitely cause an oops
      dev->next = (Scull_Dev*)MYKMALLOCPRINTK(sizeof(Scull_Dev));
      if (!dev->next) {
        return NULL;     // violates postcondition, but considered an infeasible path b/c of my model of malloc..
      }

      memset(dev->next, 0, sizeof(Scull_Dev));

      // establish invariants for the new object
      Scull_Dev *obj = dev->next;
      obj->quantum = scull_quantum;
      obj->qset = scull_qset;
      obj->prev = dev;
      obj->first = false;

      // check invariants explicitly for both modified objects
      // (this isn't needed, and could be deletedz)
      assert(all_of {
        0 < obj < objct;
        tag[obj] == Scull_Dev_tag;

        obj != dev;
        obj->prev == dev;
        !isToplevelDevice(obj);
      });
      assert(Scull_Dev_invariants(dev, mem, tag, objct, scull_devices, scull_nr_devs));
      assert(Scull_Dev_invariants(obj, mem, tag, objct, scull_devices, scull_nr_devs));
    }
    dev = dev->next;
    continue;
  }
  return dev;
}
#endif // DO_SCULL_FOLLOW


#if DO_SCULL_LSEEK
// called in response to lseek() syscall
loff_t scull_lseek(struct file *filp, loff_t off, int whence)
{
  Scull_Dev *dev = filp->private_data;
  long newpos;

  dprintkpid(DF_SYSCALL, SCULL "lseek, off=%ld, whence=%d\n",
             (long)off, whence);
  check_nonblock_change(filp, "lseek");

  switch(whence) {
    case 0: /* SEEK_SET */
      newpos = off;
      break;

    case 1: /* SEEK_CUR */
      newpos = filp->f_pos + off;
      break;

    case 2: /* SEEK_END */
      newpos = dev->totsize + off;
      break;

    default: /* can't happen */
      return -EINVAL;
  }
  if (newpos<0) return -EINVAL;

  filp->f_pos = newpos;
  return newpos;
}
#endif // DO_SCULL_LSEEK


#if DO_USERPTR_EXPERIMENT
// experiments with user pointers
void userptr_experiment(void *userptr)
{
  int size=4;
  int value;
  int err;

  dprintkpid(DF_ACCESS, SCULL "experimenting with address %p\n", userptr);

  // access_ok
  err = access_ok(VERIFY_READ, userptr, size);
  dprintkpid(DF_ACCESS, SCULL "access_ok(read): %d (1=ok, 0=not_ok)\n", err);

  // get_user
  err = get_user(value, (int*)userptr);
  dprintkpid(DF_ACCESS, SCULL "get_user: %d (0=ok, <0=not_ok)\n", err);

  // __get_user
  err = __get_user(value, (int*)userptr);
  dprintkpid(DF_ACCESS, SCULL "__get_user: %d (0=ok, <0=not_ok)\n", err);

  // copy_from_user
  err = copy_from_user(&value, userptr, size);
  dprintkpid(DF_ACCESS, SCULL "copy_from_user: %d (0=ok, >0=not_ok)\n", err);

  // __copy_from_user
  err = __copy_from_user(&value, userptr, size);
  dprintkpid(DF_ACCESS, SCULL "__copy_from_user: %d (0=ok, >0=not_ok)\n", err);
}
#endif // DO_USERPTR_EXPERIMENT


#if DO_SCULL_IOCTL
// called in response to the ioctl() syscall
Err scull_ioctl(struct inode *inode, struct file *filp,
                unsigned int cmd, unsigned long arg)
{
  Scull_Dev *dev = filp->private_data;
  bool err = 0;

  // extract the fields of 'cmd'
  int direction = _IOC_DIR(cmd);
  int magic = _IOC_TYPE(cmd);
  int cmdCode = _IOC_NR(cmd);
  int size = _IOC_SIZE(cmd);

  check_nonblock_change(filp, "ioctl");

  dprintkpid(DF_SYSCALL, SCULL "ioctl, code=%d dir=%d magic=%d code=%d size=%d\n",
             cmd, direction, magic, cmdCode, size);

  // sanity-check them; I'm not checking the size or direction,
  // THAT MIGHT BE A PROBLEM because the user can construct the
  // code such that the 'cmdCode' is valid but the 'direction'
  // and 'size' fields are such that we bypass access_ok.....!
  // update: but since we switch on the full 'cmd' below, such
  // a command will simply be reported as invalid a little later..
  // so the worst case is the user tricks us into calling access_ok,
  // but there shouldn't be any negative consequence to that
  if (!( magic == SCULL_IOC_MAGIC &&
         cmdCode >= SCULL_MIN_CODE &&
         cmdCode <= SCULL_MAX_CODE )) {
    goto invalid_code;
  }

  // if the command will access userspace, check that the userspace
  // pointer is valid
  if (direction & _IOC_READ) {
    // _IOC_READ is from the user's perspective; VERIFY_WRITE is from
    // the kernel's perspective; they both refer to kernel -> user xfer
    if (!access_ok(VERIFY_WRITE, (void*)arg, size)) {
      dprintkpid(DF_ACCESS, SCULL "access_ok(write) reports INVALID address %p\n",
                 (void*)arg);
      //return -EFAULT;
    }
    else {
      dprintkpid(DF_ACCESS, SCULL "access_ok(write) reports valid address %p\n",
                 (void*)arg);
    }
  }
  if (direction & _IOC_WRITE) {
    if (!access_ok(VERIFY_READ, (void*)arg, size)) {
      dprintkpid(DF_ACCESS, SCULL "access_ok(read) reports INVALID address %p\n",
                 (void*)arg);
      //return -EFAULT;
    }
    else {
      dprintkpid(DF_ACCESS, SCULL "access_ok(read) reports valid address %p\n",
                 (void*)arg);
    }
  }

  // at this point we assume (wrongly, as indicated above) that all
  // user pointers are valid and we can get or set them at will; but
  // since we switch on the full 'cmd', and not simply 'cmdCode', a
  // user-constructed command will hit the default: tag below
  switch (cmd) {
    case SCULL_MALLOC_STATS:
      //printAllocStats("scull");
      break;

    case SCULL_STORE_VALUE:
      userptr_experiment((void*)arg);
      userptr_experiment((void*)0);
      userptr_experiment((void*)0x70000000);
      userptr_experiment((void*)0x10000000);
      userptr_experiment((void*)0xc0000000);

      err = get_user(dev->ioctl_scratch_value, (int*)arg);
      if (err > 0) {
        printkpid(SCULL_ERR, "get_user says invalid address %p\n", (void*)arg);
        return -EFAULT;
      }
      dprintkpid(DF_SCRATCH, SCULL "got value %d from user\n",
                 dev->ioctl_scratch_value);
      break;

    case SCULL_STORE_VALUE_UNCHECKED:
      err = __get_user(dev->ioctl_scratch_value, (int*)arg);
      if (err > 0) {
        printkpid(SCULL_ERR, "__get_user says invalid address %p\n", (void*)arg);
        return -EFAULT;
      }
      dprintkpid(DF_SCRATCH, SCULL "got value %d from user\n",
                 dev->ioctl_scratch_value);
      break;

    case SCULL_GET_VALUE:
      dprintkpid(DF_SCRATCH, SCULL "giving %d to user\n",
                 dev->ioctl_scratch_value);
      err = put_user(dev->ioctl_scratch_value, (int*)arg);
      if (err > 0) {
        printkpid(SCULL_ERR, "put_user says invalid address %p\n", (void*)arg);
        return -EFAULT;
      }
      break;

    default:
      // this catches codes constructed by the user that bypass the
      // checks at the top
      dprintkpid(DF_BADIOCTL, SCULL "2nd catch invalid:\n");

    invalid_code:
      dprintkpid(DF_BADIOCTL,
                 SCULL "invalid ioctl code %d: dir=%d magic=%d code=%d size=%d\n",
                 cmd, direction, magic, cmdCode, size);
      return -EINVAL;
  }

  // success
  return 0;
}
#endif // DO_SCULL_IOCTL


#if DO_CHECK_NONBLOCK_CHANGE
void check_nonblock_change(struct file *filp, char const *caller)
{
  Scull_Dev *dev = filp->private_data;

  // which way is 'filp' set?
  int filpNonblock = (filp->f_flags & O_NONBLOCK)? 1 : 0;

  // is this different from what it was last time?
  if (filpNonblock != dev->nonblocking) {
    // record the new value
    dev->nonblocking = filpNonblock;

    // print a message
    dprintkpid(DF_NONBLOCK, SCULL "in %s, detected change of nonblock, new value: %d\n",
               caller, dev->nonblocking);
  }
}
#endif // DO_CHECK_NONBLOCK_CHANGE


#if DO_INIT_MODULE
// called when 'insmod' loads the module
Err init_module(void)
  pre(module_state==MS_UNLOADED)
  post(all_of {
    result <= 0;
    result==0 ==> module_state==MS_LOADED;
    result<0  ==> module_state==MS_UNLOADED;
  })
  changes(mem, meta, tag, objsize, objct, module_state, scull_devices, scull_major,
          Scull_Dev_tag /*all Scull fields*/)
{
  Err resULT, i;
  Scull_Dev *temp_devices;

  // sanity check the parameters
  if (scull_nr_devs < 1 || scull_nr_devs > 4) {
    printk(SCULL_ERR "scull_nr_devs must be in range [1,4]; you supplied %d.\n",
           scull_nr_devs);
    // annoying: even though I want to return EINVAL, it gets translated
    // to EBUSY by the time insmod reports it... that sucks
    return -EINVAL;
  }

  /*
   * Register your major, and accept a dynamic number
   */
  resULT = register_chrdev(scull_major, "scull", &scull_fops);
  if (resULT < 0) {
    printk(SCULL_ERR "can't get major %d\n", scull_major);
    return resULT;
  }
  if (scull_major == 0) {
    // we asked for a dynamic major node, and got it
    scull_major = resULT; /* dynamic */
  }

  // now print banner since scull_major is set
  printk(SCULL_INFO "initing; MAJ=%d devs=%d qntm=%d qset=%d debug=%s\n",
         scull_major, scull_nr_devs, scull_quantum, scull_qset, debug);

  // if we have to deallocate 'scull_devices' because of a failure,
  // then the verifier will complain I could have caused some 
  // 'filp->private_data' to dangle.  the right solution is to maintain
  // a reference set in Scull_Dev, but for now I'm going to try noticing
  // that any such filp's cannot point past the current allocation
  // boundary
  int old_objct = objct;
  assert(
    forall(file *filp). tag[filp]==file_tag ==> all_of {
      // if open and associated with this module, then the private_data
      // points below the allocation boundary
      filp->state==FS_OPEN && filp->id==SCULL_MODULE_ID ==>
        0 < filp->private_data < old_objct;
    }
  );

  /*
   * allocate the devices -- we can't have them static, as the number
   * can be specified at load time
   */
  // can't assign scull_devices yet because state is MS_UNLOADED, so
  // it's supposed to be NULL; and setting to MS_LOADED wouldn't work
  // b/c then all the tags would have to be Scull_Dev_tag
  temp_devices = (Scull_Dev*)MYKMALLOCPRINTK(scull_nr_devs * sizeof (Scull_Dev));
  if (!temp_devices) {
    resULT = -ENOMEM;
    goto fail_malloc;
  }

  // initialize the devices; in particular, this sets the 'size' fields
  // to 0 and the 'next' and 'data' fields to NULL
  memset(temp_devices, 0, scull_nr_devs * sizeof (Scull_Dev));
  mem = vc_expression(updbothranges(mem,       // manual model for now
    temp_devices, temp_devices+scull_nr_devs,    // address range
    0, Scull_Dev_size,                           // index range
    0));

  // mark all the new devices as being in the zero state
  tag = vc_expression(updflatrange(tag, temp_devices, temp_devices+scull_nr_devs,
                                   ZEROED_SCULL_DEV_TAG));

  for (i=0; i < scull_nr_devs; i++) {
    invariant {
      global_inv;

      0 <= i < scull_nr_devs;

      0 < old_objct <= temp_devices < temp_devices+scull_nr_devs <= objct;

      forall(file *filp). tag[filp]==file_tag ==> all_of {
        // if open and associated with this module, then the private_data
        // points below the allocation boundary
        filp->state==FS_OPEN && filp->id==SCULL_MODULE_ID ==>
          0 < filp->private_data < old_objct;
      };

      // first half of array
      forall(int j). 0<=j<i ==> all_of {
        tag[temp_devices+j] == Scull_Dev_tag;
        (temp_devices+j)->prev == NULL;
        (temp_devices+j)->next == NULL;
        (temp_devices+j)->first == true;
      };

      // second half of array
      forall(int k). i<=k<scull_nr_devs ==> all_of {
        tag[temp_devices+k] == ZEROED_SCULL_DEV_TAG;
        (temp_devices+k)->next == NULL;
        (temp_devices+k)->prev == NULL;
        (temp_devices+k)->data == NULL;
      };
    }

    // have to set the tag first, so that the subsequent operations
    // will find that tag
    tag = vc_expression(tag{temp_devices+i := Scull_Dev_tag});

    temp_devices[i].quantum = scull_quantum;
    temp_devices[i].qset = scull_qset;
    temp_devices[i].first = true;
  }

  // now we can claim that 'scull_devices' meets its MS_LOADED requirements
  scull_devices = temp_devices;
  module_state = MS_LOADED;
  invariant {
    module_state==MS_LOADED;
    global_inv;

    // old filps point below the old allocation boundary
    old_objct <= scull_devices;
    forall(file *filp). tag[filp]==file_tag ==> all_of {
      // if open and associated with this module, then the private_data
      // points below the allocation boundary
      filp->state==FS_OPEN && filp->id==SCULL_MODULE_ID ==>
        0 < filp->private_data < old_objct;
    };

    // all the 'next' pointers are NULL; this is needed in case I
    // need to deallocate them, since otherwise the verifier thinks
    // I might leave someone's 'prev' to dangle
    forall(int j). 0<=j<scull_nr_devs ==> all_of {
      (scull_devices+j)->next == NULL;
    };
  }

  /* At this point call the init function for any friend device */
  //if ( (resULT = init_pipe_module()) > 0 ) {    // BUG!
  if ( (resULT = init_pipe_module()) < 0 ) {
    goto fail_pipe;
  }

  // I'm normally going to have DEBUG true
  #ifndef DEBUG
    // old: register_symtab(NULL); /* otherwise, leave global symbols visible */
    EXPORT_NO_SYMBOLS();
  #endif

  #ifdef SCULL_USE_PROC /* only when available */
    /* this is the last line in init_module */
    proc_register_dynamic(&proc_root, &scull_proc_entry);
  #endif

  return 0; /* succeed */

fail_pipe:
  assert(module_state==MS_LOADED && global_inv);
  mykfreesz(scull_devices+0, scull_nr_devs * sizeof(Scull_Dev));
  scull_devices = NULL;
  module_state = MS_UNLOADED;

  {
    Scull_Dev *dev;
    assume(tag[dev]==Scull_Dev_tag);
    assert(Scull_Dev_invariants(dev, mem, tag, objct, scull_devices, scull_nr_devs));
  }


fail_malloc:
  unregister_chrdev(scull_major, "scull");

  return resULT;
}
#endif // DO_INIT_MODULE


#if DO_CLEANUP_MODULE
// called when the module is being unloaded by 'rmmod'
void cleanup_module(void)
  pre(all_of {
    module_state==MS_LOADED;
    
    // this is a prerequisite of calling cleanup_module, though
    // the kernel might find it hard to prove, in which case the
    // kernel author may decide to just 'assume' it
    forall(file *filp). !(
      tag[filp]==file_tag &&
      filp->state==FS_OPEN &&     // no open files of mine
      filp->id==SCULL_MODULE_ID
    );
  })
  post(module_state==MS_UNLOADED)
  changes(mem, meta, tag, objsize, objct, module_state, scull_devices,
          Scull_Dev_tag /*all Scull fields*/)
{
  int i;
  Err resULT;

  resULT = unregister_chrdev(scull_major, "scull");
  if (resULT < 0) {
    // but what can I do about it?
    printk(SCULL_ERR "unregister_chrdev failed with code %d!\n", resULT);
  }

  #ifdef SCULL_USE_PROC
    proc_unregister(&proc_root, scull_proc_entry.low_ino);
  #endif

  for (i=0; i<scull_nr_devs; i++) {
    invariant {
      // precondition
      global_inv;
      module_state==MS_LOADED;
      forall(file *filp). !(
        tag[filp]==file_tag &&
        filp->state==FS_OPEN &&     // no open files of mine
        filp->id==SCULL_MODULE_ID
      );
                     
      // this loop
      0 <= i < scull_nr_devs;

      // left half of the array has been trimmed
      forall(int j). 0<=j<i ==> (scull_devices+j)->next == NULL;
    }

    if (scull_trim(scull_devices+i) < 0) {
      assert(forall(int j). 0<=j<i ==> (scull_devices+j)->next == NULL);

      // there is no way to cancel an unload..
      printk(KERN_CRIT "scull: device being unloaded even though there are "
             "still ongoing operations.. this is bad..\n");

      // this will leak, but I can accept that (in fact this should not
      // happen b/c for a node to be locked a file must still be open,
      // but cleanup_module isn't supposed to be called then)
      (scull_devices+i)->next = NULL;
      assume(!exists(Scull_Dev *dev). dev->prev == scull_devices+i);
    }
  }

  // entire array has been trimmed
  assert(forall(int j). 0<=j<scull_nr_devs ==>
    (scull_devices+j)->next == NULL);

  // not going to cause any private_data to dangle
  assert(
    forall(file *filp). !(
      tag[filp]==file_tag &&
      filp->state==FS_OPEN &&     // no open files of mine
      filp->id==SCULL_MODULE_ID
    )
  );

  mykfreesz(scull_devices+0, scull_nr_devs * sizeof (Scull_Dev));
  scull_devices = NULL;
  module_state = MS_UNLOADED;

  /* and call the cleanup functions for friend devices */
  cleanup_pipe_module();

  //if (debugFlagIsSet(DF_LEAKS)) {
  //  printAllocStats("scull");
  //} 

  printk(SCULL_INFO "device unloaded.\n");
}
#endif // DO_CLEANUP_MODULE


// EOF
