; state.thy
; symbols for accessing and interpreting object state fields
; requires: memory.thy

(DEFPRED (isDead obj))
(DEFPRED (isLive obj))

(BG_PUSH (AND

  ; ------------ symbolic constants ---------
               
  ; selection offset within objects where the state resides
  (EQ state -1)                                            

  ; a selection offset for a value, when there's only one
  (EQ value 0)

  ; state value meaning the value is dead
  (EQ dead -1)
  
  ; one possible state value meaning live
  (EQ live 0)

  ; --------- literal memories ----------

  ; 0 is special: selecting any field yields zero
  (FORALL (i)
    (EQ (sel 0 i) 0))

  ; dead is also special, it's dead throughout
  (FORALL (i)
    (EQ (sel dead i) dead))

  ; shorthand for a memory with a given state and value
  (FORALL (x y)
    (EQ (sel (stateValue x y) state) x))
  (FORALL (x y)
    (EQ (sel (stateValue x y) 0) y))

  (FORALL (v)       (PATS (liveValue v))         
    (EQ (liveValue v)
        (stateValue live v)))

  ; ---------- live/dead tests --------------
  ; isDead: true if the value is dead (not to be used)
  (FORALL (obj)
    (IFF
      (isDead obj)
      (EQ (sel obj state) dead)
    ))
    
  ; isLive: true if the value can be used
  (FORALL (obj)
    (IFF
      (isLive obj)
      (NEQ (sel obj state) dead)
    ))

  ; -------------- value access --------------
  ; (valueOf v): get the value part (as opposed to the field part)
  (FORALL (v)    (PATS (valueOf v))
    (EQ
      (valueOf v)
      (sel v value)
    ))

))
