// vml_build.h
// convenience routines for building VML AST nodes

#ifndef VML_BUILD_H
#define VML_BUILD_H

#include "vml.ast.gen.h"      // VML AST

Pred *p_and(Pred *p1, Pred *p2);
Pred *p_or(Pred *p1, Pred *p2);
Pred *p_or3(Pred *p1, Pred *p2, Pred *p3);
Pred *p_impl(Pred *p1, Pred *p2);
Pred *p_not(Pred *p);

Pred *p_compare(VExpr *e1, ComparisonOp op, VExpr *e2);
Pred *p_compare3(VExpr *e1, ComparisonOp op12, VExpr *e2,
                            ComparisonOp op23, VExpr *e3);
Pred *p_compare4(VExpr *e1, ComparisonOp op12, VExpr *e2,
                            ComparisonOp op23, VExpr *e3,
                            ComparisonOp op34, VExpr *e4);

// 0 <= val < limit
Pred *p_bounds(VExpr *val, VExpr *limit);

Pred *p_equal(VExpr *e1, VExpr *e2);
Pred *p_notEqual(VExpr *e1, VExpr *e2);

Pred *p_forall1(SourceLoc loc, StringRef label, StringRef quantVarName,
                VExpr * /*nullable*/ patspec, Pred *body);


// negate 'e', which has not been given to anything else,
// so it's ok to modify it
VExpr *e_neg(VExpr *e);

VExpr *e_lit(int i);
VExpr *e_plus(VExpr *e1, VExpr *e2);

ASTList<VExpr> *exprlist(VExpr *e1, VExpr *e2, VExpr *e3);
ASTList<VExpr> *exprlist(VExpr *e1, VExpr *e2);
ASTList<VExpr> *exprlist(VExpr *e1);

Identifier *makeId(StringRef s, SourceLoc loc = SL_INIT);

VE_apply *apply1(Identifier *func, VExpr *e1);
VE_apply *apply2(Identifier *func, VExpr *e1, VExpr *e2);
VE_apply *apply3(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3);
VE_apply *apply4(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3, VExpr *e4);
VE_apply *apply5(Identifier *func, VExpr *e1, VExpr *e2, VExpr *e3, VExpr *e4, VExpr *e5);

#endif // VML_BUILD_H
