// vml_lexer.cc            see license.txt for copyright and terms of use
// code for vml_lexer.h

#include "vml_lexer.h"        // VMLLexer

char const *toString(VMLToken type)
{
  static char const * const map[] = {
    "<eof>",

    "<intlit>",
    "<id>",
    "<stringlit>",

    "__pat",
    "AND",
    "axiom",
    "assert",
    "assume",
    "const",
    "declare_partial_order",
    "else",
    "exists",
    "false",
    "forall",
    "funcsym",
    "goto",
    "if",
    "invariant",
    "OR",
    "post",
    "pre",
    "procedure",
    "pure_assert",
    "reads",
    "standalone_invariant",
    "true",
    "var",
    "while",
    "writes",

    "{",
    "}",
    ";",
    ":",
    ":=",
    "(",
    ")",
    ".",
    "!",
    "&&",
    "||",
    "==>",
    "<==>",
    "==",
    "!=",
    "<",
    "<=",
    ">",
    ">=",
    "+",
    "-",
    "*",
    "[",
    "]",
    ",",
    "->",
    "=",
    "?",

    "<highprec>",
    "<locprec>"
  };
  ASSERT_TABLESIZE(map, NUM_VML_TOKENS);
  
  xassert((unsigned)type < NUM_VML_TOKENS);
  return map[type];
}


VMLLexer::VMLLexer(StringTable &strtable, char const *fname)
  : BaseLexer(strtable, fname)
{}

VMLLexer::~VMLLexer()
{}


int VMLLexer::svalTok(int t)
{
  updLoc();
  sval = (SemanticValue)addString(yytext, yyleng);
  return t;
}


string VMLLexer::tokenDesc() const
{   
  switch (type) {
    case TOK_INT_LIT:
      return stringc << "int(" << (int)sval << ")";
      
    case TOK_IDENTIFIER:
      return stringc << "id(" << (StringRef)sval << ")";
      
    default:
      return toString((VMLToken)type);
  }
}

string VMLLexer::tokenKindDesc(int kind) const
{
  return toString((VMLToken)kind);
}


// ------------------ test code -----------------
#ifdef TEST_VML_LEXER

#include "exc.h"        // xBase
#include "srcloc.h"     // SourceLocManager

int main(int argc, char **argv)
{
  if (argc < 2) {
    cout << "usage: " << argv[0] << " input.vml\n";
    return 0;
  }

  try {
    SourceLocManager mgr;
    StringTable table;
    VMLLexer lexer(table, argv[1]);
    lexer.tokenFunc(&lexer);

    while (lexer.type != TOK_EOF) {
      cout << toString(lexer.loc) << ": " << lexer.tokenDesc() << "\n";
      lexer.getTokenFunc()(&lexer);
    }

    return 0;
  }
  catch (xBase &x) {
    cout << x << endl;
    return 4;
  }
}

#endif // TEST_VML_LEXER
